import tempfile
import unittest

import xarray as xr

from bluemath_tk.downloaders.ecmwf.ecmwf_downloader import ECMWFDownloader


class TestECMWFDownloader(unittest.TestCase):
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()
        self.downloader = ECMWFDownloader(
            product="OpenData",
            base_path_to_download="OpenDataJavixu",  # self.temp_dir,
            check=False,  # Just check paths to download, do not actually download
        )

    def test_list_datasets(self):
        datasets = self.downloader.list_datasets()
        self.assertIsInstance(datasets, list)
        self.assertTrue(len(datasets) > 0)
        print(f"Available datasets: {datasets}")

    def test_download_data(self):
        dataset = self.downloader.download_data(
            load_data=False,
            param=["msl"],
            step=[0, 240],
            type="fc",
            force=False,
        )
        self.assertIsInstance(dataset, xr.Dataset)
        print(dataset)


if __name__ == "__main__":
    unittest.main()



import logging
import warnings
from datetime import datetime, timedelta
from typing import Any, Dict

import cartopy.crs as ccrs
import cartopy.feature as cfeature
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray as xr
from matplotlib.axes import Axes
from matplotlib.collections import Collection

from ..core.dask import setup_dask_client
from ..core.decorators import validate_data_xwt
from ..core.models import BlueMathModel
from ..core.pipeline import BlueMathPipeline
from ..core.plotting.colors import get_cluster_colors, get_config_variables
from ..datamining.kma import KMA
from ..datamining.pca import PCA

warnings.filterwarnings("ignore")
config_variables = get_config_variables()


def get_dynamic_estela_predictor(
    data: xr.Dataset,
    estela: xr.Dataset,
    check_interpolation: bool = True,
    verbose: bool = False,
) -> xr.Dataset:
    """
    Transform an xarray dataset of longitude, latitude, and time into one where
    each longitude, latitude value at each time is replaced by the corresponding
    time - t, where t is specified in the estela dataset.

    Parameters
    ----------
    data : xr.Dataset
        The input dataset with dimensions longitude, latitude, and time.
    estela : xr.Dataset
        The dataset containing the F values with dimensions longitude and latitude.
    check_interpolation : bool, optional
        Whether to check if the data is interpolated. Default is True.
    verbose : bool, optional
        Whether to print verbose output. Default is False.
        If False, Dask logs are suppressed.
        If True, Dask logs are shown.

    Returns
    -------
    xr.Dataset
        The transformed dataset.
    """

    if not verbose:
        # Suppress Dask logs
        logging.getLogger("distributed").setLevel(logging.ERROR)
        logging.getLogger("distributed.client").setLevel(logging.ERROR)
        logging.getLogger("distributed.scheduler").setLevel(logging.ERROR)
        logging.getLogger("distributed.worker").setLevel(logging.ERROR)
        logging.getLogger("distributed.nanny").setLevel(logging.ERROR)
        # Also suppress bokeh and tornado logs that Dask uses
        logging.getLogger("bokeh").setLevel(logging.ERROR)
        logging.getLogger("tornado").setLevel(logging.ERROR)

    # TODO: Add customization for dask client
    _dask_client = setup_dask_client(n_workers=4, memory_limit=0.25)

    if check_interpolation:
        if (
            "longitude" not in data.dims
            or "latitude" not in data.dims
            or "time" not in data.dims
        ):
            raise ValueError("Data must have longitude, latitude, and time dimensions.")
        if "longitude" not in estela.dims or "latitude" not in estela.dims:
            raise ValueError("Estela must have longitude and latitude dimensions.")
        estela = estela.interp_like(data)  # TODO: Check NaNs interpolation

    data = data.chunk({"time": 365}).where(estela.F >= 0.0, np.nan)
    estela_traveltimes = estela.where(estela.F >= 0, np.nan).traveltime.astype(int)
    estela_max_traveltime = estela_traveltimes.max().values

    for traveltime in range(estela_max_traveltime):
        data = data.where(estela_traveltimes != traveltime, data.shift(time=traveltime))

    return data.compute()


def check_model_is_fitted(func):
    """
    Decorator to check if the model is fitted.
    """

    def wrapper(self, *args, **kwargs):
        if self.kma_bmus is None:
            raise XWTError("Fit the model before calling this property.")
        return func(self, *args, **kwargs)

    return wrapper


class XWTError(Exception):
    """Custom exception for XWT class."""

    def __init__(self, message="XWT error occurred."):
        self.message = message
        super().__init__(self.message)


class XWT(BlueMathModel, BlueMathPipeline):
    """
    Xly Weather Types (XWT) class.

    This class implements the XWT method to identify and classify weather patterns
    in a dataset. The XWT method is a combination of Principal Component Analysis (PCA)
    and K-means clustering (KMA).

    Attributes
    ----------
    steps : Dict[str, BlueMathModel]
        The steps of the XWT method.
    num_clusters : int
        The number of clusters.
    kma_bmus : pd.DataFrame
        The KMA best matching units (BMUs).
    """

    def __init__(self, steps: Dict[str, BlueMathModel]) -> None:
        """
        Initialize the XWT.

        Parameters
        ----------
        steps : Dict[str, BlueMathModel]
            The steps of the XWT method. The steps must include a PCA and a KMA model.
        """

        super().__init__()
        self.set_logger_name(name=self.__class__.__name__, level="INFO")

        # Save XWT attributes
        if steps:
            if (
                not all(isinstance(step, BlueMathModel) for step in steps.values())
                or "pca" not in steps
                or "kma" not in steps
            ):
                raise XWTError("The steps must include a PCA and a KMA model.")
        self.steps = steps
        self._data: xr.Dataset = None
        self.num_clusters: int = None
        self.kma_bmus: pd.DataFrame = None

        # Exclude attributes from being saved
        self._exclude_attributes = ["_data"]

    @property
    def data(self) -> xr.Dataset:
        return self._data

    @property
    @check_model_is_fitted
    def clusters_probs_df(self) -> pd.DataFrame:
        """
        Calculate the probabilities for each XWT.
        """

        # Calculate probabilities for each cluster
        clusters_probs = (
            self.kma_bmus["kma_bmus"].value_counts(normalize=True).sort_index()
        )

        return clusters_probs

    @property
    @check_model_is_fitted
    def clusters_monthly_probs_df(self) -> pd.DataFrame:
        """
        Calculate the monthly probabilities for each XWT.
        """

        # Calculate probabilities for each month
        monthly_probs = (
            self.kma_bmus.groupby(self.kma_bmus.index.month)["kma_bmus"]
            .value_counts(normalize=True)
            .unstack()
            .fillna(0)
        )

        return monthly_probs

    @property
    @check_model_is_fitted
    def clusters_seasonal_probs_df(self) -> pd.DataFrame:
        """
        Calculate the seasonal probabilities for each XWT.
        """

        # Calculate probabilities for each season
        # Define seasons: DJF (Dec, Jan, Feb), MAM (Mar, Apr, May),
        # JJA (Jun, Jul, Aug), SON (Sep, Oct, Nov)
        seasons = {
            "DJF": [12, 1, 2],
            "MAM": [3, 4, 5],
            "JJA": [6, 7, 8],
            "SON": [9, 10, 11],
        }
        # Add a 'season' column to the DataFrame
        kma_bmus_season = self.kma_bmus.copy()
        kma_bmus_season["season"] = kma_bmus_season.index.month.map(
            lambda x: next(season for season, months in seasons.items() if x in months)
        )

        # Calculate probabilities for each season
        seasonal_probs = (
            kma_bmus_season.groupby("season")["kma_bmus"]
            .value_counts(normalize=True)
            .unstack()
            .fillna(0)
        )

        return seasonal_probs

    @property
    @check_model_is_fitted
    def clusters_annual_probs_df(self) -> pd.DataFrame:
        """
        Calculate the annual probabilities for each XWT.
        """

        # Calculate probabilities for each year
        annual_probs = (
            self.kma_bmus.groupby(self.kma_bmus.index.year)["kma_bmus"]
            .value_counts(normalize=True)
            .unstack()
            .fillna(0)
        )

        return annual_probs

    @property
    @check_model_is_fitted
    def clusters_perpetual_year_probs_df(self) -> pd.DataFrame:
        """
        Calculate the perpetual year probabilities for each XWT.
        """

        # Calculate probabilities for each natural day in the year
        natural_day_probs = (
            self.kma_bmus.groupby(self.kma_bmus.index.dayofyear)["kma_bmus"]
            .value_counts(normalize=True)
            .unstack()
            .fillna(0)
        )
        # Set index to be the datetime first day of month
        natural_day_probs.index = [
            datetime(2000, 1, 1) + timedelta(days=i - 1)
            for i in natural_day_probs.index
        ]

        return natural_day_probs

    @property
    @check_model_is_fitted
    def get_conditioned_probabilities(self) -> pd.DataFrame:
        """
        Calculate conditional probabilities P(X_t = j | X_{t-lag} = i)
        """

        # Convert to numpy array if not already
        data = self.kma_bmus.values.flatten()

        # Find unique values in the data
        unique_values = np.unique(data)

        # Create empty matrix for conditional probabilities
        cond_probs = np.zeros((self.num_clusters, self.num_clusters))

        # Count transitions
        for i in range(len(data) - 1):
            prev_idx = np.where(unique_values == data[i])[0][0]
            next_idx = np.where(unique_values == data[i + 1])[0][0]
            cond_probs[prev_idx, next_idx] += 1

        # Normalize to get probabilities
        row_sums = cond_probs.sum(axis=1, keepdims=True)
        # Avoid division by zero
        row_sums[row_sums == 0] = 1
        cond_probs = cond_probs / row_sums

        # Create DataFrame with labels
        df_cond_probs = pd.DataFrame(
            cond_probs,
            index=[f"Cluster {v}" for v in unique_values],
            columns=[f"Cluster {v}" for v in unique_values],
        )

        return df_cond_probs

    @validate_data_xwt
    def fit(
        self,
        data: xr.Dataset,
        fit_params: Dict[str, Dict[str, Any]] = {},
        variable_to_sort_bmus: str = None,
    ) -> None:
        """
        Fit the XWT model to the data.

        Parameters
        ----------
        data : xr.Dataset
            The data to fit the model to. Must be PCA formatted.
        fit_params : Dict[str, Dict[str, Any]], optional
            The fitting parameters for the PCA and KMA models. Default is {}.
        variable_to_sort_bmus : str, optional
            The variable to sort the BMUs. Default is None.

        Raises
        ------
        XWTError
            If the data is not PCA formatted.

        TODO: Standarize PCs by first PC variance.
              pca.pcs_df / pca.pcs.stds.isel(n_component=0).values ??
        """

        # Make a copy of the data to avoid modifying the original dataset
        self._data = data.copy()

        pca: PCA = self.steps.get("pca")
        if pca.pcs is None:
            try:
                _pcs_ds = pca.fit_transform(
                    data=data,
                    **fit_params.get("pca", {}),
                )
            except Exception as e:
                raise XWTError(f"Error during PCA fitting: {e}")
        else:
            self.logger.info("PCA already fitted, skipping PCA fitting.")

        kma: KMA = self.steps.get("kma")
        self.num_clusters = kma.num_clusters

        data_to_kma = pca.pcs_df.copy()

        if "regression_guided" in fit_params.get("kma", {}):
            guiding_vars = fit_params["kma"]["regression_guided"].get("vars", [])

            if guiding_vars:
                guiding_data = pd.DataFrame(
                    {var: data[var].values for var in guiding_vars},
                    index=data.time.values,
                )
                data_to_kma = pd.concat([data_to_kma, guiding_data], axis=1)

        kma_bmus, _kma_bmus_df = kma.fit_predict(
            data=data_to_kma,
            **fit_params.get("kma", {}),
        )
        self.kma_bmus = kma_bmus + 1  # TODO: Check if this is necessary!!!

        # Re-sort kma clusters based on variable if specified
        if variable_to_sort_bmus:
            pca.pcs["kma_bmus"] = (("time"), self.kma_bmus["kma_bmus"].values)
            sorted_bmus = (
                pca.inverse_transform(
                    PCs=pca.pcs.groupby("kma_bmus")
                    .mean()
                    .rename({"kma_bmus": pca.pca_dim_for_rows})
                )
                .mean(dim=pca.coords_to_stack)
                .sortby(variable_to_sort_bmus)[f"{pca.pca_dim_for_rows}"]
                .values
            )
            sorted_bmus_mapping = dict(
                zip(sorted_bmus, range(1, self.num_clusters + 1))
            )
            self.kma_bmus.replace(sorted_bmus_mapping, inplace=True)

        # Add the KMA bmus to the PCs and data
        pca.pcs["kma_bmus"] = (("time"), self.kma_bmus["kma_bmus"].values)
        self.data["kma_bmus"] = (("time"), self.kma_bmus["kma_bmus"].values)

    def plot_map_features(
        self, ax: Axes, land_color: str = cfeature.COLORS["land"]
    ) -> None:
        """
        Plot map features on an axis.

        Parameters
        ----------
        ax : Axes
            The axis to plot the map features on.
        land_color : str, optional
            The color of the land. Default is cfeature.COLORS["land"].
        """

        ax.add_feature(cfeature.COASTLINE)
        ax.add_feature(cfeature.BORDERS, linestyle=":")
        ax.add_feature(cfeature.LAND, edgecolor="black", color=land_color)
        ax.add_feature(cfeature.OCEAN, color="lightblue", alpha=0.5)

    def plot_xwts(
        self, var_to_plot: str, anomaly: bool = False, map_center: tuple = None
    ) -> Collection:
        """
        Plot the XWTs for a variable.

        Parameters
        ----------
        var_to_plot : str
            The variable to plot.
        anomaly : bool, optional
            Whether to plot the anomaly of the variable. Default is False.
        map_center : tuple, optional
            The center of the map. Default is None.

        Returns
        -------
        GridSpec
            The grid specification with the XWTs plot.
        """

        if anomaly:
            data_to_plot = self.data.groupby("kma_bmus").mean()[
                var_to_plot
            ] - self.data[var_to_plot].mean("time")
        else:
            data_to_plot = self.data.groupby("kma_bmus").mean()[var_to_plot]

        if self.num_clusters > 3:
            col_wrap = int(np.ceil(np.sqrt(self.num_clusters)))
        else:
            col_wrap = self.num_clusters

        # Get the configuration for the variable to plot if it exists
        var_to_plot_config = config_variables.get(var_to_plot, {})
        # Get the cluster colors for each XWT
        xwts_colors = get_cluster_colors(num_clusters=self.num_clusters)

        # Create figure with enough space at bottom for colorbar
        fig = plt.figure(figsize=(15, 16))
        gs = gridspec.GridSpec(
            col_wrap,
            col_wrap,
            wspace=0.05,
            hspace=0.05,
        )

        # Plot the XWTs for the variable
        vmin = var_to_plot_config.get("vmin", data_to_plot.min().values)
        vmax = var_to_plot_config.get("vmax", data_to_plot.max().values)
        for i, (bmus, xwt_color) in enumerate(
            zip(data_to_plot.kma_bmus.values, xwts_colors)
        ):
            row = i // col_wrap
            col = i % col_wrap
            if map_center:
                ax = fig.add_subplot(
                    gs[row, col], projection=ccrs.Orthographic(*map_center)
                )
                p = data_to_plot.sel(kma_bmus=bmus).plot(
                    ax=ax,
                    cmap=var_to_plot_config.get("cmap", "RdBu_r"),
                    add_colorbar=False,
                    transform=ccrs.PlateCarree(),
                    vmin=vmin,
                    vmax=vmax,
                )
                self.plot_map_features(ax=ax, land_color=xwt_color)
            else:
                ax = fig.add_subplot(gs[row, col])
                p = data_to_plot.sel(kma_bmus=bmus).plot(
                    ax=ax,
                    cmap=var_to_plot_config.get("cmap", "RdBu_r"),
                    add_colorbar=False,
                    vmin=vmin,
                    vmax=vmax,
                )
                for border in ["top", "bottom", "left", "right"]:
                    ax.spines[border].set_color(xwt_color)
            ax.set_title("")
            ax.text(
                0.05,
                0.05,
                int(bmus),
                ha="left",
                va="bottom",
                fontsize=15,
                fontweight="bold",
                color="navy",
                transform=ax.transAxes,
            )

        # Add colorbar in little custom axes at the bottom
        cbar_ax = fig.add_axes([0.3, 0.05, 0.4, 0.02])
        _cb = fig.colorbar(
            p,
            cax=cbar_ax,
            orientation="horizontal",
            label=var_to_plot_config.get("label", var_to_plot),
        )

        return p

    def _axplot_wt_probs(
        self,
        ax: Axes,
        wt_probs: np.ndarray,
        ttl: str = "",
        vmin: float = 0.0,
        vmax: float = 0.1,
        cmap: str = "Blues",
        caxis: str = "black",
        plot_text: bool = False,
    ) -> Collection:
        """
        Axes plot WT cluster probabilities.

        Parameters
        ----------
        ax : Axes
            The axis to plot the WT cluster probabilities on.
        wt_probs : np.ndarray
            The WT cluster probabilities.
        ttl : str, optional
            The title of the plot. Default is "".
        vmin : float, optional
            The minimum value of the colorbar. Default is 0.0.
        vmax : float, optional
            The maximum value of the colorbar. Default is 0.1.
        cmap : str, optional
            The colormap to use. Default is "Blues".
        caxis : str, optional
            The color of the axis. Default is "black".
        plot_text : bool, optional
            Whether to plot the text in each cell. Default is False.
        """

        # cluster transition plot
        pc = ax.pcolor(
            np.flipud(wt_probs),
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            edgecolors="k",
        )
        # plot text in each cell
        if plot_text:
            for i in range(wt_probs.shape[0]):
                for j in range(wt_probs.shape[1]):
                    ax.text(
                        j + 0.5,
                        wt_probs.shape[0] - 0.5 - i,
                        f"{wt_probs[i, j]:.2f}",
                        ha="center",
                        va="center",
                        fontsize=6,
                        fontweight="bold",
                        color="black",
                    )

        # customize axes
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_title(ttl, {"fontsize": 10, "fontweight": "bold"})

        # axis color
        plt.setp(ax.spines.values(), color=caxis)
        plt.setp(
            [ax.get_xticklines(), ax.get_yticklines()],
            color=caxis,
        )

        # axis linewidth
        if caxis != "black":
            plt.setp(ax.spines.values(), linewidth=3)

        return pc

    def _axplot_wt_hist(self, ax: Axes, ttl: str = "") -> Axes:
        """
        Axes plot WT cluster count histogram.

        Parameters
        ----------
        ax : Axes
            The axis to plot the WT cluster count histogram on.
        ttl : str, optional
            The title of the plot. Default is "".

        Returns
        -------
        Axes
            The axis with the WT cluster count histogram.
        """

        # cluster transition plot
        ax.hist(
            self.kma_bmus.values.reshape(-1),
            bins=np.arange(1, self.num_clusters + 2),
            edgecolor="k",
        )

        # customize axes
        # ax.grid('y')

        ax.set_xticks(np.arange(1, self.num_clusters + 1) + 0.5)
        ax.set_xticklabels(np.arange(1, self.num_clusters + 1))
        ax.set_xlim([1, self.num_clusters + 1])
        ax.tick_params(axis="both", which="major", labelsize=6)
        ax.set_title(ttl, {"fontsize": 10, "fontweight": "bold"})

        return ax

    def plot_dwts_probs(
        self,
        vmax: float = 0.15,
        vmax_seasonality: float = 0.15,
        plot_text: bool = False,
    ) -> None:
        """
        Plot Daily Weather Types bmus probabilities.

        Parameters
        ----------
        vmax : float, optional
            The maximum value of the colorbar. Default is 0.15.
        vmax_seasonality : float, optional
            The maximum value of the colorbar for seasonality. Default is 0.15.
        plot_text : bool, optional
            Whether to plot the text in each cell. Default is False.

        Raises
        ------
        ValueError
            If the kma_bmus time sampling is not daily.
        """

        if (self.kma_bmus.index[-1] - self.kma_bmus.index[-2]) != timedelta(days=1):
            raise ValueError("The kma_bmus time sampling must be daily.")

        # Best rows cols combination
        if self.num_clusters > 3:
            n_rows = n_cols = int(np.ceil(np.sqrt(self.num_clusters)))
        else:
            n_cols = self.num_clusters
            n_rows = 1

        # figure
        fig = plt.figure(figsize=(15, 9))
        gs = gridspec.GridSpec(4, 7, wspace=0.10, hspace=0.25)

        # list all plots params
        l_months = [
            (1, "January", gs[1, 3]),
            (2, "February", gs[2, 3]),
            (3, "March", gs[0, 4]),
            (4, "April", gs[1, 4]),
            (5, "May", gs[2, 4]),
            (6, "June", gs[0, 5]),
            (7, "July", gs[1, 5]),
            (8, "August", gs[2, 5]),
            (9, "September", gs[0, 6]),
            (10, "October", gs[1, 6]),
            (11, "November", gs[2, 6]),
            (12, "December", gs[0, 3]),
        ]
        l_3months = [
            ([12, 1, 2], "DJF", gs[3, 3]),
            ([3, 4, 5], "MAM", gs[3, 4]),
            ([6, 7, 8], "JJA", gs[3, 5]),
            ([9, 10, 11], "SON", gs[3, 6]),
        ]

        # plot total probabilities
        c_T = self.clusters_probs_df.values
        C_T = c_T.reshape(n_rows, n_cols)
        ax_probs_T = plt.subplot(gs[:2, :2])
        pc = self._axplot_wt_probs(
            ax_probs_T, C_T, ttl="DWT Probabilities", plot_text=plot_text
        )

        # plot counts histogram
        ax_hist = plt.subplot(gs[2:, :3])
        _ax_hist = self._axplot_wt_hist(ax_hist, ttl="DWT Counts")

        # plot probabilities by month
        for m_ix, m_name, m_gs in l_months:
            try:
                c_M = self.clusters_monthly_probs_df.loc[m_ix, :].values
                C_M = c_M.reshape(n_rows, n_cols)
                ax_M = plt.subplot(m_gs)
                self._axplot_wt_probs(
                    ax_M, C_M, ttl=m_name, vmax=vmax, plot_text=plot_text
                )
            except Exception as e:
                self.logger.error(e)

        # plot probabilities by 3 month sets
        for m_ix, m_name, m_gs in l_3months:
            try:
                c_M = self.clusters_seasonal_probs_df.loc[m_name, :].values
                C_M = c_M.reshape(n_rows, n_cols)
                ax_M = plt.subplot(m_gs)
                self._axplot_wt_probs(
                    ax_M,
                    C_M,
                    ttl=m_name,
                    vmax=vmax_seasonality,
                    cmap="Greens",
                    plot_text=plot_text,
                )
            except Exception as e:
                self.logger.error(e)

        # add custom colorbar
        pp = ax_probs_T.get_position()
        cbar_ax = fig.add_axes([pp.x1 + 0.02, pp.y0, 0.02, pp.y1 - pp.y0])
        cb = fig.colorbar(pc, cax=cbar_ax, cmap="Blues")
        cb.ax.tick_params(labelsize=8)

    def plot_perpetual_year(self) -> Axes:
        """
        Plot perpetual year bmus probabilities.

        Returns
        -------
        Axes
            The plot with the perpetual year bmus probabilities.
        """

        # Get cluster colors for stacked bar plot
        cluster_colors = get_cluster_colors(self.num_clusters)
        cluster_colors_list = [
            tuple(cluster_colors[cluster, :]) for cluster in range(self.num_clusters)
        ]

        # Plot perpetual year bmus
        fig, ax = plt.subplots(1, figsize=(15, 5))
        clusters_perpetual_year_probs_df = self.clusters_perpetual_year_probs_df
        clusters_perpetual_year_probs_df.plot.area(
            ax=ax,
            stacked=True,
            color=cluster_colors_list,
            legend=False,
        )
        ax.set_ylim(0, 1)

        return ax



import itertools
from copy import deepcopy
from typing import Any, Callable, Dict, List, Union

import numpy as np
import pandas as pd
import xarray as xr


class BlueMathPipeline:
    """
    A flexible, modular pipeline for chaining together BlueMath models and data processing steps.

    This class allows you to define a sequence of steps, where each step must be a BlueMathModel.
    Each step is defined by a dictionary specifying:
        - 'name': str, a unique identifier for the step.
        - 'model': the model instance to use (or will be created via 'model_init' and 'model_init_params').
        - 'model_init': (optional) a callable/class to instantiate the model.
        - 'model_init_params': (optional) dict of parameters for model initialization.
        - 'fit_method': (optional) str, the method name to call for fitting (default is based on model type).
        - 'fit_params': (optional) dict, parameters for the fit method.
        - 'pipeline_attributes_to_store': (optional) list of attribute names to store for later use.

    The pipeline supports advanced parameter passing, including referencing outputs from previous steps
    and using callables for dynamic parameter computation.
    """

    # Map model class names to their default fit method
    _default_fit_methods = {
        "LHS": "generate",
        "MDA": "fit_predict",
        "KMA": "fit_predict",
        "SOM": "fit_predict",
        "PCA": "fit_transform",
        "RBF": "fit_predict",
    }

    def __init__(self, steps: List[Dict[str, Any]]):
        """
        Initialize the BlueMathPipeline with a sequence of steps.

        Parameters
        ----------
        steps : List[Dict[str, Any]]
            A list of dictionaries, each specifying at least 'name' and 'model',
            and optionally 'model_init', 'model_init_params', 'fit_method', 'fit_params',
            and 'pipeline_attributes_to_store'.
        """

        self.steps = steps
        self._pipeline_attributes = {}  # Stores attributes from previous models for later reference

    @property
    def pipeline_attributes(self) -> Dict[str, Dict[str, Any]]:
        """
        Get the stored model attributes from each pipeline step.

        Returns
        -------
        Dict[str, Dict[str, Any]]
            A dictionary mapping step names to dictionaries of stored attributes.

        Raises
        ------
        ValueError
            If the pipeline has not been fit yet and no attributes are stored.
        """

        if len(self._pipeline_attributes) == 0:
            raise ValueError(
                "No model attributes found. Please fit the pipeline first."
            )

        return self._pipeline_attributes

    def fit(self, data: Union[np.ndarray, pd.DataFrame, xr.Dataset] = None):
        """
        Fit all models in the pipeline sequentially, passing the output of each step as input to the next.

        For each step, the model is (optionally) initialized, then fit using the specified method and parameters.
        Parameters and model initialization arguments can be dynamically computed using callables or references
        to previous pipeline attributes.

        Parameters
        ----------
        data : Union[np.ndarray, pd.DataFrame, xr.Dataset], optional
            The input data to fit the models. If None, the pipeline expects each step to handle its own data.

        Returns
        -------
        The output of the final step in the pipeline (could be transformed data, predictions, etc.).
        """

        if data is not None:
            data = deepcopy(data)  # Avoid modifying the original input data

        # Iterate over each step in the pipeline
        for step in self.steps:
            # If model needs to be initialized (using model_init and model_init_params)
            if "model_init" in step and "model_init_params" in step:
                for init_param_name, init_param_value in step[
                    "model_init_params"
                ].items():
                    # If the parameter is a callable, call it with (pipeline, step, data)
                    if callable(init_param_value):
                        step["model_init_params"][init_param_name] = init_param_value(
                            self, step, data
                        )
                    # If the parameter is a dict with 'data' and 'function', call the function
                    elif (
                        isinstance(init_param_value, dict)
                        and "data" in init_param_value
                        and "function" in init_param_value
                        and callable(init_param_value["function"])
                    ):
                        # Call the function with (pipeline, step, data)
                        step["model_init_params"][init_param_name] = init_param_value[
                            "function"
                        ](self, step, init_param_value["data"])
                    # If the parameter is the string 'data', replace with the current data
                    elif isinstance(init_param_value, str):
                        if init_param_value == "data":
                            step["model_init_params"][init_param_name] = data

                # Actually instantiate the model with the resolved parameters
                step["model"] = step["model_init"](**step["model_init_params"])

            # Retrieve the model instance for this step
            model = step["model"]
            default_method = self._default_fit_methods.get(type(model).__name__)
            method_name = step.get(
                "fit_method", default_method
            )  # Use step's method or default
            if method_name is None:
                raise ValueError(
                    f"No fit method found for model {type(model).__name__}. Please specify a fit_method in the step."
                )

            # Prepare parameters for the fit method, resolving any callables or references
            params = step.get("fit_params", {}).copy()
            for param_name, param_value in params.items():
                if callable(param_value):
                    params[param_name] = param_value(self, step, data)
                elif (
                    isinstance(param_value, dict)
                    and "data" in param_value
                    and "function" in param_value
                    and callable(param_value["function"])
                ):
                    # Call the function with (data, pipeline, step)
                    params[param_name] = param_value["function"](
                        self, step, param_value["data"]
                    )
                elif isinstance(param_value, str):
                    if param_value == "data":
                        params[param_name] = data

            # Call the fit method on the model with the resolved parameters
            method = getattr(model, method_name)
            try:
                # Some methods expect 'data' as a named argument
                data = method(data=data, **params)
            except Exception as _e:
                # If that fails, try calling without 'data' as a named argument
                data = method(**params)

            # Store specified model attributes for later use, if requested
            if "pipeline_attributes_to_store" in step:
                self._pipeline_attributes[step["name"]] = {
                    attr_name: getattr(model, attr_name)
                    for attr_name in step["pipeline_attributes_to_store"]
                }

        return data

    def _generate_param_combinations(
        self, param_grid: Dict[str, List[Any]]
    ) -> List[Dict[str, Any]]:
        """
        Generate all possible combinations of parameters from a parameter grid for a single pipeline step.

        Parameters
        ----------
        param_grid : Dict[str, List[Any]]
            Dictionary mapping parameter names to lists of values to try for each parameter.

        Returns
        -------
        List[Dict[str, Any]]
            List of dictionaries, each representing a unique combination of parameters.
        """

        keys = param_grid.keys()
        values = param_grid.values()
        combinations = list(itertools.product(*values))

        return [dict(zip(keys, combo)) for combo in combinations]

    def grid_search(
        self,
        data: Union[np.ndarray, pd.DataFrame],
        param_grid: List[Dict[str, Any]],
        metric: Callable = None,
        target_data: Union[np.ndarray, pd.DataFrame] = None,
        plot: bool = False,
    ) -> Dict[str, Any]:
        """
        Perform a grid search over all possible parameter combinations for all steps in the pipeline.

        This method evaluates every possible combination of parameters (from the provided grids) for each step,
        fits the pipeline, and scores the result using the provided metric or the last model's score method.
        The best parameter set (lowest score) is selected and the pipeline is updated accordingly.

        Parameters
        ----------
        data : Union[np.ndarray, pd.DataFrame]
            The input data to fit the models.
        param_grid : List[Dict[str, Any]]
            List of parameter grids for each step in the pipeline. Each grid is a dict mapping parameter names
            to lists of values to try. Parameters can be for either model_init_params or fit_params.
        metric : Callable, optional
            Function to evaluate the final output. Should take (y_true, y_pred) as arguments.
            If None, will use the last model's built-in score method if available.
        target_data : Union[np.ndarray, pd.DataFrame], optional
            Target data to evaluate against if using a custom metric. Required if metric is provided.
        plot : bool, optional
            If True, plot the score for each parameter combination after grid search. Default is False.

        Returns
        -------
        Dict[str, Any]
            Dictionary containing:
                - 'best_params': the best parameter set for each step
                - 'best_score': the best score achieved
                - 'best_output': the output of the pipeline for the best parameters
                - 'all_results': a list of all parameter sets and their scores/outputs

        Raises
        ------
        ValueError
            If the number of parameter grids does not match the number of pipeline steps,
            or if a metric is provided but no target_data is given.
        """

        if len(param_grid) != len(self.steps):
            raise ValueError(
                "Number of parameter grids must match number of pipeline steps"
            )

        if metric is not None and target_data is None:
            raise ValueError("target_data must be provided when using a custom metric")

        # Generate all possible parameter combinations for each step
        all_param_combinations = []
        for step_params in param_grid:
            step_combinations = self._generate_param_combinations(step_params)
            all_param_combinations.append(step_combinations)

        # Cartesian product: all possible combinations across all steps
        param_combinations = list(itertools.product(*all_param_combinations))

        best_score = float("inf")  # Initialize best score as infinity (lower is better)
        best_params = None
        best_output = None
        all_results = []

        # Iterate over every possible parameter combination
        for step_params in param_combinations:
            pipeline_copy = deepcopy(self)  # Work on a copy to avoid side effects

            # Update each step in the pipeline with the current parameter set
            for step_idx, params in enumerate(step_params):
                step = pipeline_copy.steps[step_idx]
                # Assign parameters to model_init_params or fit_params as appropriate
                if "model_init_params" in step:
                    for param_name, param_value in params.items():
                        if param_name in step["model_init_params"]:
                            step["model_init_params"][param_name] = param_value
                        else:
                            step.setdefault("fit_params", {})[param_name] = param_value
                else:
                    step.setdefault("fit_params", {}).update(params)
                # Re-initialize the model if model_init_params were updated
                if "model_init_params" in step and "model_init" in step:
                    step["model"] = step["model_init"](**step["model_init_params"])

            # Fit the pipeline and get the output for this parameter set
            output = pipeline_copy.fit(data)

            # Score the output using the provided metric or the model's score method
            if metric is not None:
                score = metric(target_data, output)
            else:
                try:
                    score = pipeline_copy.steps[-1]["model"].score(target_data, output)
                except (AttributeError, TypeError):
                    raise ValueError(
                        "Either provide a metric function and target_data, "
                        "or ensure the last model has a score method"
                    )

            # Store the result for this parameter set
            result = {"params": step_params, "score": score, "output": output}
            all_results.append(result)

            # Update the best score/params/output if this is the best so far
            if score < best_score:
                best_score = score
                best_params = step_params
                best_output = output

        # After search, update the pipeline with the best parameters found
        for step_idx, params in enumerate(best_params):
            step = self.steps[step_idx]
            if "model_init_params" in step:
                for param_name, param_value in params.items():
                    if param_name in step["model_init_params"]:
                        step["model_init_params"][param_name] = param_value
                    else:
                        step.setdefault("fit_params", {})[param_name] = param_value
            else:
                step.setdefault("fit_params", {}).update(params)
            if "model_init_params" in step and "model_init" in step:
                step["model"] = step["model_init"](**step["model_init_params"])

        # Plotting if requested
        if plot:
            try:
                import matplotlib.pyplot as plt

                scores = [result["score"] for result in all_results]
                plt.figure(figsize=(6, 4))
                plt.plot(range(len(scores)), scores, marker="o", linestyle="-")
                plt.xlabel("Parameter Combination Index")
                plt.ylabel("Score")
                plt.title("Grid Search Scores for Parameter Combinations")
                plt.grid(True)
                plt.show()
            except ImportError:
                print("matplotlib is not installed. Cannot plot grid search results.")

        return {
            "best_params": best_params,
            "best_score": best_score,
            "best_output": best_output,
            "all_results": all_results,
        }



import os

import numpy as np
from scipy.interpolate import interp1d

from .._base_wrappers import BaseModelWrapper

sbatch_file_example = """#!/bin/bash
#SBATCH --ntasks=1              # Number of tasks (MPI processes)
#SBATCH --partition=geocean     # Standard output and error log
#SBATCH --nodes=1               # Number of nodes to use
#SBATCH --mem=4gb               # Memory per node in GB (see also --mem-per-cpu)
#SBATCH --time=24:00:00


case_dir=$(ls | awk "NR == $SLURM_ARRAY_TASK_ID")
cd $case_dir
launchSchism.sh
"""


class SchismModelWrapper(BaseModelWrapper):
    """
    Wrapper for the Schism model.

    Attributes
    ----------
    default_parameters : dict
        The default parameters type for the wrapper.
    available_launchers : dict
        The available launchers for the wrapper.
    """

    default_parameters = {}

    available_launchers = {"geoocean-cluster": "launchSchism.sh"}

    def __init__(
        self,
        templates_dir: str,
        metamodel_parameters: dict,
        fixed_parameters: dict,
        output_dir: str,
        templates_name: dict = "all",
        debug: bool = True,
    ) -> None:
        """
        Initialize the Schism model wrapper.
        """

        super().__init__(
            templates_dir=templates_dir,
            metamodel_parameters=metamodel_parameters,
            fixed_parameters=fixed_parameters,
            output_dir=output_dir,
            templates_name=templates_name,
            default_parameters=self.default_parameters,
        )
        self.set_logger_name(
            name=self.__class__.__name__, level="DEBUG" if debug else "INFO"
        )

        self.sbatch_file_example = sbatch_file_example

    def build_case(
        self,
        case_context: dict,
        case_dir: str,
    ) -> None:
        """
        Build the input files for a case.

        Parameters
        ----------
        case_context : dict
            The case context.
        case_dir : str
            The case directory.
        """

        os.makedirs(os.path.join(case_dir, "outputs"), exist_ok=True)


class SchismHyCoFfEE(SchismModelWrapper):
    """
    Schism wrapper: HyCoFfEE : A Hybrid Compound Flooding metamodel for Estuarine Environment.
    """

    default_parameters = {
        # .param file
        "indvel": {
            "type": float,
            "value": 1,
            "description": "Method for computing velocity at nodes: 0= linear shape function; 1 = averaging method",
        },
        "ihorcon": {
            "type": float,
            "value": 0,
            "description": "Horizontal viscosity option: 0= no viscosity is used; 1= Lapacian",
        },
        "ishapiro": {
            "type": float,
            "value": 0,
            "description": "on/off flag for Shapiro filter: 0= off; 1= on",
        },
        "inter_mom": {
            "type": float,
            "value": 0,
            "description": " Interpolation is used for velocity at foot of char: 0=linear; 1=Kriging (define kr_co ); -1= spatial Kriging (define krvel.gr3)",
        },
        # CORE
        "ipre": {
            "type": float,
            "value": 0,
            "description": "Pre-process flag: 0= normal run; 1= only for single CPU!",
        },
        "ibc": {
            "type": float,
            "value": 1,
            "description": "Barotropic flag",
        },
        "ibtp": {
            "type": float,
            "value": 0,
            "description": "Baroclinic flag",
        },
        # TRACERS
        "ntracer_gen": {
            "type": float,
            "value": 2,
            "description": "# of tracers",
        },
        "ntracer_age": {
            "type": float,
            "value": 4,
            "description": "Age calculation, must be 2*ntracer_gen",
        },
        # SEDIMENTS
        "sed_class": {
            "type": float,
            "value": 5,
            "description": "SED3D",
        },
        # ECO
        "eco_class": {
            "type": float,
            "value": 27,
            "description": "EcoSim. Must be bewteen [25,60]",
        },
        # OPT
        "ieos_type": {
            "type": float,
            "value": 0,
            "description": "UNESCO 1980 (nonlinear); =1: linear function of T ONLY, i.e. ",
        },
        "ieos_pres": {
            "type": float,
            "value": 0,
            "description": "Used only if ieos_type=0. 0: without pressure effects ",
        },
        "eos_a": {
            "type": float,
            "value": -0.1,
            "description": "Needed if ieos_type=1; should be <=0",
        },
        "eos_b": {
            "type": float,
            "value": 1001,
            "description": "Needed if ieos_type=1",
        },
        # WWM
        "msc2": {
            "type": float,
            "value": 24,
            "description": "Grid for WWM. Same as msc in .nml ... for consitency check between SCHISM and WWM",
        },
        "mdc2": {
            "type": float,
            "value": 30,
            "description": "Grid for WWM. Same as mdc in .nml ... for consitency check between SCHISM and WWM",
        },
        # SCHOUT
        "output_WL": {
            "type": float,
            "value": 1,
            "description": "Elev. [m] = 0: off; 1: on",
        },
        "output_wind_velocity": {
            "type": float,
            "value": 1,
            "description": "Wind velocity vector [m/s] = 0: off; 1: on",
        },
        "output_water_temperature": {
            "type": float,
            "value": 0,
            "description": "Water temperature [C] = 0: off; 1: on",
        },
        "output_depth_averaged_vel": {
            "type": float,
            "value": 1,
            "description": "Depth-averaged vel vector [m/s] = 0: off; 1: on",
        },
        "output_water_salinity": {
            "type": float,
            "value": 0,
            "description": "Water salinity [PSU] = 0: off; 1: on",
        },
    }

    def generate_time_series(
        self,
        case_context: dict,
        case_dir: str,
    ):
        """
        Generate the required th files for the case.

        Parameters
        ----------
        case_context : dict
            The case context.
        case_dir : str
            The case directory.
        """

        print(case_context["case_num"])

        # Calculate the start and end times in seconds
        qp = case_context.get("Qp")
        qb = case_context.get("Qb")
        tp = case_context.get("Tp")
        tau_ss_d = case_context.get("Tau_ss_d")
        tau = case_context.get("Tau")
        cm = case_context.get("CM")
        sb = case_context.get("Sb")
        sp = case_context.get("Sp")
        hidro = case_context.get("hidro")

        # Puede ser que tau(ss-d) sea negativo pero Tp sea tan pequeño que las 24 horas del surge empiecen antes que el hidrograma de caudal
        time0 = tp + abs(tau_ss_d) - 24

        # hidrograma SCS
        hidro_Qp = hidro.Q * (qp - qb) + qb
        if (
            tau_ss_d <= 0 and time0 >= 0 and tp != 0
        ):  # el hidrograma marca el inicio del tiempo
            hidro_hours = hidro.t.values * tp
        else:
            if tau_ss_d <= 0 and time0 < 0 and tp != 0:
                hidro_hours = (
                    hidro.t.values * tp + abs(time0)
                )  # el storm surge marca el inicio y hay que sumar el Tau con respecto al surge
            else:
                if tau_ss_d > 0:
                    t0 = 24 - (tp - tau_ss_d)
                    hidro_hours = (
                        hidro.t.values * tp + t0
                    )  # el storm surge marca el inicio y hay que sumar el Tau con respecto al surge
                else:  # el desfase es cero
                    if (
                        tp == 0
                    ):  # el tiempo de concentración también, con lo que se supone que el caudal es constante e igual al base
                        hidro_hours = np.arange(0, 49, 1)
                        hidro_Qp = np.ones(len(hidro_hours)) * qp

        hidro_seg = hidro_hours * 3600

        # duración teniendo en cuenta los desfases entre picos de caudal y storm surge
        # en el caso de Tau(ss-d) < 0 y Tp suficientemente grande para que el hidrograma de caudal marque el inicio
        timef = tp + abs(tau_ss_d) + 24

        # AT
        # Creamos un array con los valores de tiempo (48 horas)
        int_min = case_context.get("dt")

        if (
            timef > np.max(hidro_hours) and tau_ss_d < 0 and time0 > 0
        ):  # añadido nuevo el time0>0
            t = np.arange(0, timef * 3600, int_min)
            # interpolamos caudal a esa resolución temporal
            f = interp1d(hidro_seg, hidro_Qp, kind="linear")
            # f = interp1d(hidro_seg, hidro_Qp, kind="linear", fill_value="extrapolate", bounds_error=False) # CAMBIADO!!!!

            t2 = np.arange(0, np.max(hidro_seg), int_min)
            hidro_Qp_5_1 = f(t2)
            t3 = np.max(t) - np.max(t2)
            hidro_Qp_5_2 = np.zeros(int(t3 / int_min)) + hidro_Qp.values[-1]
            hidro_Qp_5 = np.hstack((hidro_Qp_5_1, hidro_Qp_5_2))
        else:
            if (
                hidro_hours[-1] < 48 and tau_ss_d < 0 and time0 > 0
            ):  # añadido nuevo el time0>0
                t = np.arange(0, 48 * 3600, int_min)
                # interpolamos caudal a esa resolución temporal
                f = interp1d(hidro_seg, hidro_Qp, kind="linear")
                t2 = np.arange(0, np.max(hidro_seg), int_min)
                hidro_Qp_5_1 = f(t2)
                t3 = np.max(t) - np.max(t2)
                hidro_Qp_5_2 = np.zeros(int(t3 / int_min)) + hidro_Qp.values[-1]
                hidro_Qp_5 = np.hstack((hidro_Qp_5_1, hidro_Qp_5_2))
            else:
                if tau_ss_d <= 0 and time0 < 0 and tp != 0 and hidro_hours[-1] > 48:
                    t = np.arange(0, np.max(hidro_seg), int_min)
                    hidro_Qp_5_1 = np.full(int(abs(time0 * 3600 / int_min)), qb)
                    f = interp1d(hidro_seg, hidro_Qp, kind="linear")
                    t2 = np.arange(np.min(hidro_seg), np.max(hidro_seg), int_min)
                    hidro_Qp_5_2 = f(t2)
                    hidro_Qp_5 = np.hstack((hidro_Qp_5_1, hidro_Qp_5_2))
                else:
                    if tau_ss_d <= 0 and time0 < 0 and tp != 0 and hidro_hours[-1] < 48:
                        t = np.arange(0, 48 * 3600, int_min)
                        hidro_Qp_5_1 = np.full(int(abs(time0 * 3600 / int_min)), qb)
                        f = interp1d(hidro_seg, hidro_Qp, kind="linear")
                        t2 = np.arange(np.min(hidro_seg), np.max(hidro_seg), int_min)
                        hidro_Qp_5_2 = f(t2)
                        time_left = 48 - hidro_hours[-1]
                        hidro_Qp_5_3 = np.full(int(abs(time_left * 3600 / int_min)), qb)
                        hidro_Qp_5 = np.hstack(
                            (hidro_Qp_5_1, hidro_Qp_5_2, hidro_Qp_5_3)
                        )
                    else:
                        if hidro_hours[-1] < 48:
                            t0 = 24 - (tp - tau_ss_d)
                            # timef2 = t0 + np.max()
                            t = np.arange(0, 48 * 3600, int_min)
                            hidro_Qp_5_0 = np.zeros(int(t0 * 3600 / int_min)) + qb
                            # interpolamos caudal a esa resolución temporal
                            f = interp1d(hidro_seg, hidro_Qp, kind="linear")
                            t2 = np.arange(
                                np.min(hidro_seg), np.max(hidro_seg), int_min
                            )
                            hidro_Qp_5_1 = f(t2)
                            t3 = np.max(t) - np.max(t2)
                            hidro_Qp_5_2 = np.zeros(int(t3 / int_min)) + qb
                            hidro_Qp_5 = np.hstack(
                                (hidro_Qp_5_0, hidro_Qp_5_1, hidro_Qp_5_2)
                            )
                        else:
                            t = np.arange(0, np.max(hidro_seg), int_min)
                            # interpolamos caudal a esa resolución temporal
                            f = interp1d(hidro_seg, hidro_Qp, kind="linear")
                            if tau_ss_d > 0:
                                t0 = 24 - (tp - tau_ss_d)
                                # hidro_Qp_5_1 = np.full(int(df_v.iloc[n]['Tau_ss_d']*3600/5),qb)
                                hidro_Qp_5_1 = np.full(int(t0 * 3600 / int_min), qb)
                                t2 = np.arange(
                                    np.min(hidro_seg), np.max(hidro_seg), int_min
                                )
                                hidro_Qp_5_2 = f(t2)
                                hidro_Qp_5 = np.hstack((hidro_Qp_5_1, hidro_Qp_5_2))
                            else:
                                hidro_Qp_5 = f(t)  # el hidrograma abarca todo el tiempo

        # Tau (desfase de la pleamar con respecto al pico de caudal Qp - en horas)
        # tau = tau

        # AT
        at = np.sin(np.pi / 2 + np.pi * tau / 6 + 2 * np.pi * t / (12 * 3600)) * cm / 2

        # Storm surge
        time0 = tp + abs(tau_ss_d) - 24

        if tau_ss_d < 0 and (time0 > 0):  # el hidrograma marca el inicio del tiempo
            # desde el inicio del hidrograma hasta que empieza el pico, el valor del surge es el base
            # time0=df_v.iloc[n]['Tp'] + abs(df_v.iloc[n]['Tau_ss_d']) - 24
            ss0 = np.full(int(time0 * 3600 / int_min), sb)
            # 24 horas hasta el pico
            ss1 = np.linspace(sb, sp, int(24 * 3600 / int_min))
            # 24 horas después del pico
            ss2 = np.linspace(sp, sb, int(24 * 3600 / int_min))
            # lo que quede hasta el final del hidrograma
            timef2 = tp + abs(tau_ss_d) + 24
            time_final = np.max(t) - (timef2 * 3600 - int_min)
            ss3 = np.full(int(time_final / int_min), sb)
            ss = np.hstack((ss0, ss1, ss2, ss3))

        else:
            # storm surge marca el inicio del tiempo, surge centrado en las primeras 48 horas
            ss1 = np.linspace(sb, sp, int(24 * 3600 / int_min))
            ss2 = np.linspace(sp, sb, int(24 * 3600 / int_min))
            timet = time0 + hidro_hours[-1]

            if timet < 48:
                ss = np.hstack((ss1, ss2))
            else:
                # a partir de las 48 horas, el surge tiene el valor del base
                time_final = np.max(t) - (48 * 3600 - int_min)
                ss3 = np.full(int(time_final / int_min), sb)
                ss = np.hstack((ss1, ss2, ss3))

        t_inicial = np.arange(0, 4 * 24 * 3600, int_min)
        ss_inicial = np.zeros(len(t_inicial)) + sb
        at_inicial = (
            np.sin(np.pi / 2 + np.pi * tau / 6 + 2 * np.pi * t_inicial / (12 * 3600))
            * cm
            / 2
        )
        q_inicial = np.zeros(len(t_inicial)) - qb

        t_series = np.hstack((t_inicial, t_inicial[-1] + t))
        at_series = np.hstack((at_inicial, at))

        if len(ss) < len(at):
            ss4 = np.full(int(len(at) - len(ss)), ss[-1])
            ss = np.hstack((ss, ss4))
            ss_series = np.hstack((ss_inicial, ss))
        else:
            ss_series = np.hstack((ss_inicial, ss))

        if len(hidro_Qp_5) < len(at):
            qq4 = np.full(int(len(at) - len(hidro_Qp_5)), -hidro_Qp_5[-1])
            hidro_Qp_5 = np.hstack((hidro_Qp_5, qq4))
            qq_series = np.hstack((q_inicial, -hidro_Qp_5))
        else:
            qq_series = np.hstack((q_inicial, -hidro_Qp_5))

        AT_SS_series = ss_series + at_series

        return t_series, at_series, ss_series, AT_SS_series, qq_series

    def rndays_calculation(
        self,
        case_context: dict,
        case_dir: str,
        t_series: np.ndarray,
    ):
        """
        Calculate the rndays parameter (time of simulation in days).

        Parameters
        ----------
        case_context : dict
            The case context.
        case_dir : str
            The case directory.
        t_series : np.ndarray
            The time series array.
        """

        # Calculate the rndays parameter
        t_rndays = t_series[-1] / (3600 * 24)

        int_min = case_context.get("dt")  # Define the time step in seconds for .th file

        dt_nspool = case_context.get("dt") * case_context.get("nspool")

        # Generate time vector for .th file
        times_sim = np.arange(
            0,
            t_series[-1] + (case_context.get("ihfskip") / dt_nspool / 4.0) * int_min,
            int_min,
        )

        return t_rndays, times_sim

    def generate_wind_th(
        self,
        case_context: dict,
        case_dir: str,
        times: np.ndarray,
    ):
        """
        Generate the wind files for a case.

        Parameters
        ----------
        case_context : dict
            The case context.
        case_dir : str
            The case directory.
        """

        ####### PARAMETERS OBTAINED FROM metamodel_parameters #######
        wind_magnitude = case_context.get("wind_magnitude")
        wind_direction_era5 = case_context.get("wind_direction")

        wind_direction_SCHISM = (90 - wind_direction_era5) % 360
        #############################################################

        u_wind = (
            -np.ones(len(times))
            * wind_magnitude
            * np.cos(np.radians(wind_direction_SCHISM))
        )
        v_wind = (
            -np.ones(len(times))
            * wind_magnitude
            * np.sin(np.radians(wind_direction_SCHISM))
        )

        # Define the output file path inside the directory
        output_path = os.path.join(case_dir, "wind.th")

        # Create the output file
        with open(output_path, "w+") as file:
            # Write each line as time, u_wind, and v_wind separated by spaces
            for time, u, v in zip(times, u_wind, v_wind):
                file.write(f"{time} {u} {v}\n")

    def generate_at_th(
        self,
        case_context: dict,
        case_dir: str,
        at_series: np.ndarray,
        times: np.ndarray,
    ):
        """
        Generate the at.th file for a case (not required by SCHISM, but nice to plot).

        Parameters
        ----------
        case_context : dict
            The case context.
        case_dir : str
            The case directory.
        """

        # Define the output file path inside the directory
        output_path = os.path.join(case_dir, "at.th")

        # Create the output file
        with open(output_path, "w+") as file:
            # Write each line as time and at_series separated by spaces
            for time, at_serie in zip(times, at_series):
                file.write(f"{time} {at_serie}\n")

    def generate_ss_th(
        self,
        case_context: dict,
        case_dir: str,
        ss_series: np.ndarray,
        times: np.ndarray,
    ):
        """
        Generate the ss.th file for a case (not required by SCHISM, but nice to plot).

        Parameters
        ----------
        case_context : dict
            The case context.
        case_dir : str
            The case directory.
        """

        # Define the output file path inside the directory
        output_path = os.path.join(case_dir, "ss.th")

        # Create the output file
        with open(output_path, "w+") as file:
            # Write each line as time and ss_series separated by spaces
            for time, ss_serie in zip(times, ss_series):
                file.write(f"{time} {ss_serie}\n")

    def generate_elev_th(
        self,
        case_context: dict,
        case_dir: str,
        AT_SS_series: np.ndarray,
        times: np.ndarray,
    ):
        """
        Generate the elev.th file for a case.

        Parameters
        ----------
        case_context : dict
            The case context.
        case_dir : str
            The case directory.
        """

        # Define the output file path inside the directory
        output_path = os.path.join(case_dir, "elev.th")

        # Create the output file
        with open(output_path, "w+") as file:
            # Write each line as time and AT_SS_series separated by spaces
            for time, ss_serie in zip(times, AT_SS_series):
                file.write(f"{time} {ss_serie}\n")

    def generate_flux_th(
        self,
        case_context: dict,
        case_dir: str,
        qq_series: np.ndarray,
        times: np.ndarray,
    ):
        """
        Generate the elev.th file for a case.

        Parameters
        ----------
        case_context : dict
            The case context.
        case_dir : str
            The case directory.
        """

        # Define the output file path inside the directory
        output_path = os.path.join(case_dir, "flux.th")

        # Create the output file
        with open(output_path, "w+") as file:
            # Write each line as time and AT_SS_series separated by spaces
            for time, qq_serie in zip(times, qq_series):
                file.write(f"{time} {qq_serie}\n")

    def build_case(
        self,
        case_context: dict,
        case_dir: str,
    ):
        """
        Build the input files for a case.

        Parameters
        ----------
        case_context : dict
            The case context.
        case_dir : str
            The case directory.
        """

        super().build_case(case_context=case_context, case_dir=case_dir)

        # Calculate time series from the cases
        t_series, at_series, ss_series, AT_SS_series, qq_series = (
            self.generate_time_series(
                case_context=case_context,
                case_dir=case_dir,
            )
        )

        # Calculate the rndays parameter and time vector for the th files
        case_context["rnday"], times_sim = self.rndays_calculation(
            case_context=case_context,
            case_dir=case_dir,
            t_series=t_series,
        )

        # Generate the wind.th file for each case
        self.generate_wind_th(
            case_context=case_context,
            case_dir=case_dir,
            times=times_sim,
        )

        # Generate the at.th file for each case
        self.generate_at_th(
            case_context=case_context,
            case_dir=case_dir,
            at_series=at_series,
            times=times_sim,
        )

        # Generate the ss.th file for each case
        self.generate_ss_th(
            case_context=case_context,
            case_dir=case_dir,
            ss_series=ss_series,
            times=times_sim,
        )

        # Generate the ss.th file for each case
        self.generate_elev_th(
            case_context=case_context,
            case_dir=case_dir,
            AT_SS_series=AT_SS_series,
            times=times_sim,
        )

        # Generate the flux.th file for each case
        self.generate_flux_th(
            case_context=case_context,
            case_dir=case_dir,
            qq_series=qq_series,
            times=times_sim,
        )


class SchismOnlyWindModelWrapper(SchismModelWrapper):
    """
    Schism wrapper: only wind case.
    """

    default_parameters = {
        # .param file
        "indvel": {
            "type": float,
            "value": 1,
            "description": "Method for computing velocity at nodes: 0= linear shape function; 1 = averaging method",
        },
        "ihorcon": {
            "type": float,
            "value": 0,
            "description": "Horizontal viscosity option: 0= no viscosity is used; 1= Lapacian",
        },
        "ishapiro": {
            "type": float,
            "value": 0,
            "description": "on/off flag for Shapiro filter: 0= off; 1= on",
        },
        "inter_mom": {
            "type": float,
            "value": 0,
            "description": " Interpolation is used for velocity at foot of char: 0=linear; 1=Kriging (define kr_co ); -1= spatial Kriging (define krvel.gr3)",
        },
        # CORE
        "ipre": {
            "type": float,
            "value": 0,
            "description": "Pre-process flag: 0= normal run; 1= only for single CPU!",
        },
        "ibc": {
            "type": float,
            "value": 0,
            "description": "Barotropic flag",
        },
        "ibtp": {
            "type": float,
            "value": 0,
            "description": "Baroclinic flag",
        },
        # TRACERS
        "ntracer_gen": {
            "type": float,
            "value": 2,
            "description": "# of tracers",
        },
        "ntracer_age": {
            "type": float,
            "value": 4,
            "description": "Age calculation, must be 2*ntracer_gen",
        },
        # SEDIMENTS
        "sed_class": {
            "type": float,
            "value": 5,
            "description": "SED3D",
        },
        # ECO
        "eco_class": {
            "type": float,
            "value": 27,
            "description": "EcoSim. Must be bewteen [25,60]",
        },
        # OPT
        "ieos_type": {
            "type": float,
            "value": 0,
            "description": "UNESCO 1980 (nonlinear); =1: linear function of T ONLY, i.e. ",
        },
        "ieos_pres": {
            "type": float,
            "value": 0,
            "description": "Used only if ieos_type=0. 0: without pressure effects ",
        },
        "eos_a": {
            "type": float,
            "value": -0.1,
            "description": "Needed if ieos_type=1; should be <=0",
        },
        "eos_b": {
            "type": float,
            "value": 1001,
            "description": "Needed if ieos_type=1",
        },
        # WWM
        "msc2": {
            "type": float,
            "value": 24,
            "description": "Grid for WWM. Same as msc in .nml ... for consitency check between SCHISM and WWM",
        },
        "mdc2": {
            "type": float,
            "value": 30,
            "description": "Grid for WWM. Same as mdc in .nml ... for consitency check between SCHISM and WWM",
        },
        # SCHOUT
        "output_WL": {
            "type": float,
            "value": 1,
            "description": "Elev. [m] = 0: off; 1: on",
        },
        "output_wind_velocity": {
            "type": float,
            "value": 1,
            "description": "Wind velocity vector [m/s] = 0: off; 1: on",
        },
        "output_water_temperature": {
            "type": float,
            "value": 0,
            "description": "Water temperature [C] = 0: off; 1: on",
        },
        "output_depth_averaged_vel": {
            "type": float,
            "value": 1,
            "description": "Depth-averaged vel vector [m/s] = 0: off; 1: on",
        },
        "output_water_salinity": {
            "type": float,
            "value": 0,
            "description": "Water salinity [PSU] = 0: off; 1: on",
        },
    }

    def generate_windth(
        self,
        case_context: dict,
        case_dir: str,
    ):
        """
        Generate the wind files for a case.

        Parameters
        ----------
        case_context : dict
            The case context.
        case_dir : str
            The case directory.
        """

        ####### PARAMETERS OBTAINED FROM metamodel_parameters #######
        wind_magnitude = case_context.get("wind_magnitude")
        wind_direction = case_context.get("wind_direction")
        #############################################################

        total_seconds = (
            case_context.get("rndays") * 24 * 3600
        )  # Define times in seconds for .th file
        times = np.arange(
            0, total_seconds + case_context.get("dt"), case_context.get("dt")
        )  # Generate time vector
        u_wind = -[wind_magnitude * np.cos(np.radians(wind_direction))] * len(
            times
        )  # Calculate u_wind
        v_wind = -[wind_magnitude * np.sin(np.radians(wind_direction))] * len(
            times
        )  # Calculate v_wind

        # Define the output file path inside the directory
        output_path = os.path.join(case_dir, "wind.th")

        # Create the output file
        with open(output_path, "w+") as file:
            # Write each line as time, u_wind, and v_wind separated by spaces
            for time, u, v in zip(times, u_wind, v_wind):
                file.write(f"{time} {u} {v}\n")

    def build_case(
        self,
        case_context: dict,
        case_dir: str,
    ):
        """
        Build the input files for a case.

        Parameters
        ----------
        case_context : dict
            The case context.
        case_dir : str
            The case directory.
        """

        super().build_case(case_context=case_context, case_dir=case_dir)

        case_context["start_year"] = case_context.get("tini").year
        case_context["start_month"] = case_context.get("tini").month
        case_context["start_day"] = case_context.get("tini").day
        case_context["start_hour"] = case_context.get("tini").hour
        case_context["rndays"] = (
            case_context.get("tend") - case_context.get("tini")
        ).total_seconds() / 86400

        # Generate wind file for each case
        self.generate_windth(
            case_context=case_context,
            case_dir=case_dir,
        )



import math
import os
from typing import List, Union

import numpy as np
import pandas as pd
import xarray as xr

from .._base_wrappers import BaseModelWrapper


class XBeachModelWrapper(BaseModelWrapper):
    """
    Wrapper for the XBeach model.
    https://xbeach.readthedocs.io/en/latest/

    Attributes
    ----------
    default_parameters : dict
        The default parameters type for the wrapper.
    available_launchers : dict
        The available launchers for the wrapper.
    """

    default_parameters = {
        "comptime": {
            "type": int,
            "value": 3600,
            "description": "The computational time.",
        },
        "wbctype": {
            "type": str,
            "value": "off",
            "description": "The time step for the simulation.",
        },
    }

    available_launchers = {
        "geoocean-cluster": "launchXbeach.sh",
    }

    def __init__(
        self,
        templates_dir: str,
        metamodel_parameters: dict,
        fixed_parameters: dict,
        output_dir: str,
        templates_name: dict = "all",
        debug: bool = True,
    ) -> None:
        """
        Initialize the XBeach model wrapper.
        """

        super().__init__(
            templates_dir=templates_dir,
            metamodel_parameters=metamodel_parameters,
            fixed_parameters=fixed_parameters,
            output_dir=output_dir,
            templates_name=templates_name,
            default_parameters=self.default_parameters,
        )
        self.set_logger_name(
            name=self.__class__.__name__, level="DEBUG" if debug else "INFO"
        )

    def build_case(
        self,
        case_context: dict,
        case_dir: str,
    ) -> None:
        """
        Build the input files for a case.

        Parameters
        ----------
        case_context : dict
            The case context.
        case_dir : str
            The case directory.
        """

        if case_context["wbctype"] == "jonstable":
            with open(f"{case_dir}/jonswap.txt", "w") as f:
                for _i in range(math.ceil(case_context["comptime"] / 3600)):
                    f.write(
                        f"{case_context['Hs']} {case_context['Tp']} {case_context['Dir']} 3.300000 30.000000 3600.000000 1.000000 \n"
                    )

    def _get_average_var(self, case_nc: xr.Dataset, var: str) -> np.ndarray:
        """
        Get the average value of a variable except for the first hour of the simulation

        Parameters
        ----------
        case_nc : xr.Dataset
            Simulation .nc file.
        var : str
            Variable of interest.

        Returns
        -------
        np.ndarray
            The average value of the variable.
        """

        if var in case_nc:
            return np.mean(
                case_nc[var]
                .isel(meantime=slice(1, int(case_nc.meantime.values[-1])))
                .values,
                axis=0,
            )

    def _get_max_var(self, case_nc: xr.Dataset, var: str) -> np.ndarray:
        """
        Get the Max value of a variable except for the first hour of the simulation

        Parameters
        ----------
        case_nc : xr.Dataset
            Simulation .nc file.
        var : str
            Variable of interest.

        Returns
        -------
        np.ndarray
            The max value of the variable.
        """

        if var in case_nc:
            return np.max(
                case_nc[var]
                .isel(meantime=slice(1, int(case_nc.meantime.values[-1])))
                .values,
                axis=0,
            )

    def monitor_cases(self, value_counts: str = None) -> Union[pd.DataFrame, dict]:
        """
        Monitor the cases based on different model log files.
        """

        cases_status = {}

        for case_dir in self.cases_dirs:
            case_dir_name = os.path.basename(case_dir)
            if os.path.exists(os.path.join(case_dir, "XBlog.txt")):
                if (
                    os.path.exists(os.path.join(case_dir, "XBerror.txt"))
                    & os.path.getsize(os.path.join(case_dir, "XBerror.txt"))
                    != 0
                ):
                    cases_status[case_dir_name] = "XBerror.txt"
                    continue
                else:
                    with open(os.path.join(case_dir, "XBlog.txt"), "r") as f:
                        lines = f.readlines()[-2:]

                    if any("End of program xbeach" in line.lower() for line in lines):
                        cases_status[case_dir_name] = "End of run"
                        continue
                    else:
                        cases_status[case_dir_name] = "Running"
                        continue
            else:
                cases_status[case_dir_name] = "No run"
                continue

        return super().monitor_cases(
            cases_status=cases_status, value_counts=value_counts
        )

    def postprocess_case(
        self,
        case_num: int,
        case_dir: str,
        output_vars: List[str] = None,
        overwrite_output: bool = True,
    ) -> xr.Dataset:
        """
        Convert tab output files to netCDF file.

        Parameters
        ----------
        case_num : int
            The case number.
        case_dir : str
            The case directory.
        output_vars : list, optional
            The output variables to postprocess. Default is None.
        overwrite_output : bool, optional
            Overwrite the output.nc file. Default is True.

        Returns
        -------
        xr.Dataset
            The postprocessed Dataset.
        """

        import warnings

        warnings.filterwarnings("ignore")

        self.logger.info(f"[{case_num}]: Postprocessing case {case_num} in {case_dir}.")

        output_nc_path = os.path.join(case_dir, "xboutput_postprocessed.nc")
        if not os.path.exists(output_nc_path) or overwrite_output:
            output_raw = xr.open_dataset(os.path.join(case_dir, "xboutput.nc"))

            globalx = output_raw.globalx.values
            globaly = output_raw.globaly.values
            zb = output_raw.zb.values[0]
            y = np.arange(globalx.shape[0])
            x = np.arange(globalx.shape[1])

            ds = xr.Dataset(
                {
                    "globalx": (("y", "x"), globalx),
                    "globaly": (("y", "x"), globaly),
                    "zb": (("y", "x"), zb),
                },
                coords={"y": y, "x": x},
            )

            for var in output_vars:
                if var == "zs_max":
                    maxed = self._get_max_var(case_nc=output_raw, var=var)
                    masked = xr.where(ds["zb"] > 0, np.nan, maxed)
                    ds[var] = (("y", "x"), masked.data)
                else:
                    averaged = self._get_average_var(case_nc=output_raw, var=var)
                    masked = xr.where(ds["zb"] > 0, np.nan, averaged)
                    ds[var] = (("y", "x"), masked.data)

            ds = ds.drop_vars("zb")
            ds.to_netcdf(output_nc_path)

            return ds
        else:
            self.logger.info(
                f"[{case_num}]: Reading existing xboutput_postprocessed.nc file."
            )
            output_nc = xr.open_dataset(output_nc_path)

            return output_nc

    def join_postprocessed_files(
        self, postprocessed_files: List[xr.Dataset]
    ) -> xr.Dataset:
        """
        Join postprocessed files in a single Dataset.

        Parameters
        ----------
        postprocessed_files : list
            The postprocessed files.

        Returns
        -------
        xr.Dataset
            The joined xarray.Dataset.
        """

        return xr.concat(postprocessed_files, dim="case_num")



"""
Project: BlueMath_tk
Module: bluemath_tk
Author: GeoOcean Research Group, Universidad de Cantabria
Repository: https://github.com/GeoOcean/BlueMath_tk.git
Status: Under development (Working)
"""

# Import specific modules instead of using wildcard imports
from . import (
    additive,
    config,
    core,
    datamining,
    distributions,
    downloaders,
    interpolation,
    predictor,
    risk,
    tcs,
    teslakit,
    tide,
    topo_bathy,
    waves,
    wrappers,
)

# Add __all__ variable to control what gets imported when using `from module import *`.
__all__ = [
    "additive",
    "config",
    "core",
    "datamining",
    "distributions",
    "downloaders",
    "interpolation",
    "predictor",
    "risk",
    "tcs",
    "teslakit",
    "tide",
    "topo_bathy",
    "waves",
    "wrappers",
]



import os.path as op

from bluemath_tk.wrappers.delft3d.delft3d_wrapper import Delft3dModelWrapper


class Delft3DfmModelWrapper(Delft3dModelWrapper):
    """
    Wrapper for the Delft3D model with flow mode.
    """

    def build_case(
        self,
        case_context: dict,
        case_dir: str,
    ) -> None:
        """
        Build the input files for a case.

        Parameters
        ----------
        case_context : dict
            The case context.
        case_dir : str
            The case directory.
        """

        self.copy_files(
            src=case_context.get("grid_nc_file"),
            dst=op.join(case_dir, op.basename(case_context.get("grid_nc_file"))),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common 
import os
import os.path as op
import glob
import configparser
from shutil import copyfile
import pickle
import json
from datetime import timedelta, datetime

# pip
from prettytable import PrettyTable
import numpy as np
import pandas as pd
import xarray as xr
import netCDF4 as nc

# teslakit
# from .__init__ import __version__, __author__
from .io.getinfo import description
from .io.aux_nc import StoreBugXdset
from .io.matlab import ReadTCsSimulations, ReadMatfile, ReadNakajoMats, \
ReadGowMat, ReadCoastMat, ReadEstelaMat

from .util.time_operations import xds_reindex_daily, xds_reindex_monthly, \
xds_limit_dates, xds_common_dates_daily, fast_reindex_hourly, \
fast_reindex_hourly_nsim, generate_datetimes, xds_further_dates, \
repair_times_hourly

from .mjo import MJO_Categories


# i/o

def clean_files(l_files):
    'remove files at list'
    for f in l_files:
        if op.isfile(f): os.remove(f)

def fill_metadata(xds, set_source=False):
    '''
    for each variable in xarray.Dataset xds, attributes will be set
    using resources/variables_attrs.json

    set_source - True for adding package source and institution metadata
    '''

    # read attributes dictionary
    p_resources = op.join(op.dirname(op.realpath(__file__)), 'resources')
    p_vats = op.join(p_resources, 'variables_attrs.json')
    with open(p_vats) as jf:
        d_vats = json.load(jf)

    # update dataset variables (names, units, descriptions)
    for vn in xds.variables:
        if vn.lower() in d_vats.keys():
           xds[vn].attrs = d_vats[vn.lower()]

    # set global attributes (source, institution)
    if set_source:
        xds.attrs['source'] = 'teslakit_v{0}'.format(__version__)
        #xds.attrs['institution'] = '{0}'.format(__author__)

    return xds

def save_nc(xds, p_save, safe_time=False):
    '''
    (SECURE) exports xarray.Dataset to netcdf file format.

     - fills dataset with teslakit variables and source metadata
     - avoids overwritting problems
     - set safe_time=True if dataset contains dates beyond 2262
    '''

    # TODO: save con netCDF4 y no xarray

    # add teslakit metadata to xarray.Dataset
    xds = fill_metadata(xds, set_source=True)

    # remove previous file to avoid problems
    clean_files([p_save])

    # export .nc
    # TODO: auto para date_end > 2260
    if safe_time:
        StoreBugXdset(xds, p_save)  # time dimension safe
    else:
        xds.to_netcdf(p_save, 'w')


class atdict(dict):
    'modified dictionary that works using ".key" '
    __getattr__= dict.__getitem__
    __setattr__= dict.__setitem__
    __delattr__= dict.__delitem__


class Database(object):
    'Teslakit database'

    def __init__(self, data_folder):

        self.data_folder = data_folder
        self.paths = PathControl(self.data_folder)

        self.site_name = None

    def SetSite(self, site_name):
        'Sets current site'

        self.site_name = site_name
        p_site = op.join(self.paths.p_sites, site_name)

        # check site folder
        if not op.isdir(p_site):
            print('Teslakit Site not found at at {0}'.format(p_site))

        else:
            self.paths.SetSite(self.site_name)

    def MakeNewSite(self, site_name):
        'Makes all directories and .ini files needed for a new teslakit project site'

        # TODO REVISAR / UPDATEAR

        self.site_name = site_name
        p_site = op.join(self.paths.p_sites, site_name)

        # make site folder
        if not op.isdir(p_site):
            os.makedirs(p_site)

        else:
            print('Teslakit Site already exists at {0}'.format(p_site))
            self.paths.SetSite(self.site_name)
            return

        # copy site.ini template
        for fn in ['site.ini']:
            copyfile(op.join(self.paths.p_resources, fn), op.join(p_site, fn))

        # create site subfolders
        self.paths.SetSite(self.site_name)
        for sf in self.paths.site.keys():
            p_sf = op.join(self.paths.p_site, sf)
            if not op.isdir(p_sf): os.makedirs(p_sf)

        print('Teslakit Site generated at {0}'.format(p_site))

    def CheckInputFiles(self):
        'Checks teslakit required input files availability'

        # TODO REVISAR / UPDATEAR

        # files/folder to check ()
        conf = [
            ('MJO', ['hist'], [op.isfile]),
            ('TCs', ['noaa'], [op.isfile]),
            ('SST', ['hist_pacific'], [op.isfile]),
            ('WAVES', ['spectra'], [op.isfile]),
            ('ESTELA', ['coastmat', 'estelamat', 'slp'],
             [op.isfile, op.isfile, op.isfile]),
            ('TIDE', ['mareografo_nc', 'hist_astro'], [op.isfile, op.isfile]),
            #('HYCREWW', ['rbf_coef'], [op.isdir]),
            #('NEARSHORE', ['swan_projects'], [op.isdir]),
        ]

        # get status
        l_status = []
        for k1, l_k2, l_fc in conf:
            for k2, fc in zip(l_k2, l_fc):
                fp = self.paths.site[k1][k2]
                rp = op.sep.join(fp.split(op.sep)[-4:])
                l_status.append(
                    ('{0}.{1}'.format(k1,k2), fc(fp), rp)
                )

        # print status with PrettyTable library
        tb = PrettyTable()
        tb.field_names = ["database ID", "available", "site path"]

        tb.align["database ID"] = "l"
        tb.align["available"] = "l"
        tb.align["site path"] = "l"

        for did, av, ap in l_status:
            tb.add_row([did, av, ap])

        print(tb)

    # SST

    def Load_SST(self):
        xds = xr.open_dataset(self.paths.site.SST.hist_pacific)
        xds = fill_metadata(xds)
        return xds

    def Save_SST_PCA(self, xds):
        save_nc(xds, self.paths.site.SST.pca)

    def Save_SST_KMA(self, xds):
        save_nc(xds, self.paths.site.SST.kma)

    def Save_SST_PCs_fit_rnd(self, d_PCs_fit, d_PCs_rnd):

        with open(self.paths.site.SST.d_pcs_fit, 'wb') as f:
            pickle.dump(d_PCs_fit, f, protocol=pickle.HIGHEST_PROTOCOL)

        with open(self.paths.site.SST.d_pcs_rnd, 'wb') as f:
            pickle.dump(d_PCs_rnd, f, protocol=pickle.HIGHEST_PROTOCOL)

    def Save_SST_AWT_sim(self, xds):
        save_nc(xds, self.paths.site.SST.awt_sim, safe_time=True)

    def Save_SST_PCs_sim(self, xds):

        # store yearly data
        save_nc(xds, self.paths.site.SST.pcs_sim, safe_time=True)

        # resample to daily and store
        xds_d = xds_reindex_daily(xds)
        save_nc(xds_d, self.paths.site.SST.pcs_sim_d, safe_time=True)

        # resample to monthly and store
        xds_m = xds_reindex_monthly(xds)
        save_nc(xds_m, self.paths.site.SST.pcs_sim_m, safe_time=True)

    def Load_SST_PCA(self):
        return xr.open_dataset(self.paths.site.SST.pca)

    def Load_SST_KMA(self):
        return xr.open_dataset(self.paths.site.SST.kma)

    def Load_SST_PCs_fit_rnd(self):

        with open(self.paths.site.SST.d_pcs_fit, 'rb') as f:
            d_PCs_fit = pickle.load(f)

        with open(self.paths.site.SST.d_pcs_rnd, 'rb') as f:
            d_PCs_rnd = pickle.load(f)

        return d_PCs_fit, d_PCs_rnd

    def Load_SST_AWT_sim(self):
        return xr.open_dataset(self.paths.site.SST.awt_sim, decode_times=True)

    def Load_SST_PCs_sim_d(self):
        return xr.open_dataset(self.paths.site.SST.pcs_sim_d, decode_times=True)

    def Load_SST_PCs_sim_m(self):
        return xr.open_dataset(self.paths.site.SST.pcs_sim_m, decode_times=True)

    # MJO

    def Load_MJO_hist(self):
        xds = xr.open_dataset(self.paths.site.MJO.hist)
        xds = fill_metadata(xds)
        return xds

    def Save_MJO_sim(self, xds):
        save_nc(xds, self.paths.site.MJO.sim, safe_time=True)

    def Load_MJO_sim(self):
        return xr.open_dataset(self.paths.site.MJO.sim, decode_times=True)

    # TCs

    def Load_TCs_noaa(self):
        return xr.open_dataset(self.paths.site.TCs.noaa)

    def Save_TCs_r1_hist(self, xds_tcs, xds_params):
        save_nc(xds_tcs, self.paths.site.TCs.hist_r1)
        save_nc(xds_params, self.paths.site.TCs.hist_r1_params)

    def Save_TCs_r2_hist(self, xds_tcs, xds_params):
        save_nc(xds_tcs, self.paths.site.TCs.hist_r2)
        save_nc(xds_params, self.paths.site.TCs.hist_r2_params)

    def Load_TCs_r1_hist(self):
        return xr.open_dataset(self.paths.site.TCs.hist_r1), \
                xr.open_dataset(self.paths.site.TCs.hist_r1_params)

    def Load_TCs_r2_hist(self):
        return xr.open_dataset(self.paths.site.TCs.hist_r2), \
                xr.open_dataset(self.paths.site.TCs.hist_r2_params)

    def Save_TCs_r1_sim_params(self, xds):
        save_nc(xds, self.paths.site.TCs.sim_r1_params)

    def Save_TCs_r2_sim_params(self, xds):
        save_nc(xds, self.paths.site.TCs.sim_r2_params)

    def Save_TCs_r1_mda_params(self, xds):
        save_nc(xds, self.paths.site.TCs.mda_r1_params)

    def Save_TCs_r2_mda_params(self, xds):
        save_nc(xds, self.paths.site.TCs.mda_r2_params)

    def Load_TCs_r2_mda_params(self):
        return xr.open_dataset(self.paths.site.TCs.mda_r2_params)

    def Load_TCs_r2_sim_params(self):
        return xr.open_dataset(self.paths.site.TCs.sim_r2_params)

    def Load_TCs_r2_mda_Simulations(self):
        return ReadTCsSimulations(self.paths.site.TCs.mda_r2_simulations)

    def Save_TCs_sim_r2_rbf_output(self, xds):
        save_nc(xds, self.paths.site.TCs.sim_r2_rbf_output)

    def Load_TCs_sim_r2_rbf_output(self):
        return xr.open_dataset(self.paths.site.TCs.sim_r2_rbf_output)

    def Load_TCs_Nakajo(self):
        return ReadNakajoMats(self.paths.site.TCs.nakajo_mats)

    def Save_TCs_probs_synth(self, xds):
        save_nc(xds, self.paths.site.TCs.probs_synth)

    def Load_TCs_probs_synth(self):
        return xr.open_dataset(self.paths.site.TCs.probs_synth)

    def Load_TCs_r2_mda_Simulations_hytc(self):
        return xr.open_dataset(self.paths.site.TCs.mda_r2_simulations_hytc)

    def Save_TCs_r2_mda_Simulations_hytc(self, xds):
        save_nc(xds, self.paths.site.TCs.mda_r2_simulations_hytc)

    # WAVES

    def Load_WAVES_partitions_GOW(self):
        xds = ReadGowMat(self.paths.site.WAVES.partitions_gow)
        xds = fill_metadata(xds)
        return xds

    def Load_WAVES_spectra(self):
        return xr.open_dataset(self.paths.site.WAVES.spectra)

    def Save_WAVES_partitions(self, xds):
        save_nc(xds, self.paths.site.WAVES.partitions)

    def Load_WAVES_partitions(self):
        return xr.open_dataset(self.paths.site.WAVES.partitions)

    def Save_WAVES_hist(self, xds):
        save_nc(xds, self.paths.site.WAVES.hist)

    def Load_WAVES_hist(self):
        return xr.open_dataset(self.paths.site.WAVES.hist)

    # ESTELA

    def Load_ESTELA_coast(self):
        return ReadCoastMat(self.paths.site.ESTELA.coastmat)

    def Load_ESTELA_data(self):
        return ReadEstelaMat(self.paths.site.ESTELA.estelamat)

    def Load_ESTELA_SLP(self):
        return xr.open_dataset(self.paths.site.ESTELA.slp)

    def Load_ESTELA_KMA(self):
        p_est_kma = op.join(self.paths.site.ESTELA.pred_slp, 'kma.nc')
        return xr.open_dataset(p_est_kma)

    def Save_ESTELA_DWT_sim(self, xds):
        save_nc(xds, self.paths.site.ESTELA.sim_dwt, True)

    def Load_ESTELA_DWT_sim(self):
        return xr.open_dataset(
            self.paths.site.ESTELA.sim_dwt, decode_times=True)

    # HYDROGRAMS

    def Save_MU_TAU_hydrograms(self, l_xds):

        p_mutau = self.paths.site.ESTELA.hydrog_mutau
        if not op.isdir(p_mutau): os.makedirs(p_mutau)

        for x in l_xds:
            n_store = 'MUTAU_WT{0:02}.nc'.format(x.WT)
            save_nc(x, op.join(p_mutau, n_store))

    def Load_MU_TAU_hydrograms(self):

        p_mutau = self.paths.site.ESTELA.hydrog_mutau

        # MU - TAU intradaily hidrographs for each WWT
        l_mutau_ncs = sorted(
            [op.join(p_mutau, pf) for pf in os.listdir(p_mutau) if pf.endswith('.nc')]
        )
        l_xds = [xr.open_dataset(x) for x in l_mutau_ncs]
        return l_xds

    # TIDE

    def Load_TIDE_gauge(self):
        xds = xr.open_dataset(self.paths.site.TIDE.gauge)
        xds = fill_metadata(xds)

        return xds

    def Save_TIDE_hist_astro(self, xds):
        save_nc(xds, self.paths.site.TIDE.hist_astro)

    def Load_TIDE_hist_astro(self):
        xds_at = xr.open_dataset(self.paths.site.TIDE.hist_astro)
        xds_at = fill_metadata(xds_at)

        return  xds_at

    def Save_TIDE_sim_astro(self, xds):
        save_nc(xds, self.paths.site.TIDE.sim_astro, True)


    def Load_TIDE_sim_astro(self):
        xds = xr.open_dataset(self.paths.site.TIDE.sim_astro, decode_times=True)

        # manual fix problems with hourly time
        d1 = xds.time.values[0]
        d2 = d1 + timedelta(hours=len(xds.time.values[:])-1)
        time_fix =  generate_datetimes(d1, d2, 'datetime64[h]')
        xds['time'] = time_fix

        return xds

    def Save_TIDE_hist_mmsl(self, xds):
        save_nc(xds, self.paths.site.TIDE.hist_mmsl)

    def Save_TIDE_sim_mmsl(self, xds):
        save_nc(xds, self.paths.site.TIDE.sim_mmsl, True)

    def Save_TIDE_mmsl_params(self, xds):
        save_nc(xds, self.paths.site.TIDE.mmsl_model_params, True)

    def Load_TIDE_mmsl_params(self):
        xds = xr.open_dataset(self.paths.site.TIDE.mmsl_model_params)
        return xds

    def Load_TIDE_hist_mmsl(self):
        xds = xr.open_dataset(self.paths.site.TIDE.hist_mmsl)
        return xds

    def Load_TIDE_sim_mmsl(self):
        xds = xr.open_dataset(self.paths.site.TIDE.sim_mmsl, decode_times=True)
        return xds

    def Save_TIDE_mmsl_params(self, xds):
        save_nc(xds, self.paths.site.TIDE.mmsl_model_params, True)

    def Load_TIDE_mmsl_params(self):
        xds = xr.open_dataset(self.paths.site.TIDE.mmsl_model_params)
        return xds




    # COMPLETE OFFSHORE OUTPUT 

    def Generate_HIST_Covariates(self):
        '''
        Load, fix, and resample (hourly) all historical covariates:
            AWTs, DWTs, MJO, MMSL, AT
        '''

        # load data
        AWT = self.Load_SST_KMA()  # bmus + 1
        MSL = self.Load_TIDE_hist_mmsl() # mmsl (mm)
        MJO = self.Load_MJO_hist()
        DWT = self.Load_ESTELA_KMA()  # bmus + 1
        ATD_h = self.Load_TIDE_hist_astro()

        # fix WTs id format
        AWT = xr.Dataset({'bmus': AWT.bmus + 1}, coords = {'time': AWT.time})
        DWT = xr.Dataset({'bmus': (('time',), DWT.sorted_bmus_storms + 1)},
                         coords = {'time': DWT.time.values[:]})

        # get MJO categories 
        mjo_cs, _ = MJO_Categories(MJO['rmm1'], MJO['rmm2'], MJO['phase'])
        MJO['bmus'] = (('time',), mjo_cs)

        # reindex data to hourly (pad)
        AWT_h = fast_reindex_hourly(AWT)
        MSL_h = MSL.resample(time='1h').pad()
        MJO_h = fast_reindex_hourly(MJO)
        DWT_h = fast_reindex_hourly(DWT)

        # generate time envelope for output 
        d1, d2 = xds_further_dates(
            [AWT_h, ATD_h, MSL_h, MJO_h, DWT_h, ATD_h]
        )
        ten = pd.date_range(d1, d2, freq='H')

        # generate empty output dataset 
        OUT_h = xr.Dataset(coords={'time': ten})

        # prepare data
        AWT_h = AWT_h.rename({'bmus':'AWT'})
        MJO_h = MJO_h.drop_vars(['mjo','rmm1','rmm2','phase']).rename({'bmus':'MJO'})
        MSL_h = MSL_h.drop_vars(['mmsl_median']).rename({'mmsl':'MMSL'})
        MSL_h['MMSL'] = MSL_h['MMSL'] / 1000.0  # mm to m
        DWT_h = DWT_h.rename({'bmus':'DWT'})

        # TODO: revisar esto
        ATD_h = ATD_h.drop_vars(['WaterLevels','Residual']).rename({'Predicted': 'AT'})
        #ATD_h = ATD_h.drop_vars(['observed','ntr','sigma']).rename({'predicted':'AT'})

        # combine data
        xds = xr.combine_by_coords(
            [OUT_h, AWT_h, MJO_h, MSL_h, DWT_h, ATD_h],
            fill_value = np.nan,
        )

        # repair times: round to hour and remove duplicates (if any)
        xds = repair_times_hourly(xds)

        return xds

    def Save_HIST_OFFSHORE(self, xds):
        ps = self.paths.site.HISTORICAL.offshore

        s =  SplitStorage(ps)
        s.Save(xds)

    def Load_HIST_OFFSHORE(self, vns=[], decode_times=False, use_cftime=False):
        ps = self.paths.site.HISTORICAL.offshore

        s =  SplitStorage(ps)
        return s.Load(vns=vns, decode_times=decode_times, use_cftime=use_cftime)

    def Generate_SIM_Covariates(self, total_sims=None):

        # load data
        AWT = self.Load_SST_AWT_sim()
        MSL = self.Load_TIDE_sim_mmsl()
        MJO = self.Load_MJO_sim()
        DWT = self.Load_ESTELA_DWT_sim()
        ATD_h = self.Load_TIDE_sim_astro()  # hourly data, 1 sim

        # optional select total sims
        if total_sims != None:
            AWT = AWT.isel(n_sim=slice(0, total_sims))
            MSL = MSL.isel(n_sim=slice(0, total_sims))
            MJO = MJO.isel(n_sim=slice(0, total_sims))
            DWT = DWT.isel(n_sim=slice(0, total_sims))

        # reindex data to hourly (pad)
        AWT_h = fast_reindex_hourly_nsim(AWT)
        MSL_h = fast_reindex_hourly_nsim(MSL)
        MJO_h = fast_reindex_hourly_nsim(MJO)
        DWT_h = fast_reindex_hourly_nsim(DWT)

        # common dates limits
        d1, d2 = xds_limit_dates([AWT_h, MSL_h, MJO_h, DWT_h, ATD_h])
        AWT_h = AWT_h.sel(time = slice(d1, d2))
        MSL_h = MSL_h.sel(time = slice(d1, d2))
        MJO_h = MJO_h.sel(time = slice(d1, d2))
        DWT_h = DWT_h.sel(time = slice(d1, d2))
        ATD_h = ATD_h.sel(time = slice(d1, d2))

        # copy to new dataset
        times = AWT_h.time.values[:]
        xds = xr.Dataset(
            {
                'AWT': (('n_sim','time'), AWT_h.evbmus_sims.values[:].astype(int)),
                'MJO': (('n_sim','time'), MJO_h.evbmus_sims.values[:].astype(int)),
                'DWT': (('n_sim','time'), DWT_h.evbmus_sims.values[:].astype(int)),
                'MMSL': (('n_sim','time'), MSL_h.mmsl.values[:]),
                'AT': (('time',), ATD_h.astro.values[:]),
            },
            coords = {'time': times}
        )

        return xds

    def Save_SIM_OFFSHORE(self, xds, n_sim):
        ps = self.paths.site.SIMULATION.offshore

        nm = '{0:08d}'.format(n_sim)  # sim code
        ps_sim = op.join(ps, nm)

        s =  SplitStorage(ps_sim)
        s.Save(xds, safe_time=True)

    def Load_SIM_OFFSHORE(self, n_sim, vns=[], decode_times=False, use_cftime=False):
        ps = self.paths.site.SIMULATION.offshore

        nm = '{0:08d}'.format(n_sim)  # sim code
        ps_sim = op.join(ps, nm)

        s =  SplitStorage(ps_sim)
        return s.Load(vns=vns, decode_times=decode_times, use_cftime=use_cftime)

    def Load_SIM_OFFSHORE_all(self, vns=[], decode_times=False, use_cftime=False):
        ps = self.paths.site.SIMULATION.offshore

        # locate simulations
        sims = sorted([x for x in os.listdir(ps) if x.isdigit() and op.isdir(op.join(ps, x))])

        # read all simulations
        for c, nm in enumerate(sims):
            ps_sim = op.join(ps, nm)
            s = SplitStorage(ps_sim)

            # read time and variables from first file
            if c==0:
                ds = s.Load(vns=vns, decode_times=False)

                # generate output xarray.Dataset, dims and vars
                out = xr.Dataset({}, coords={'time': ds.time, 'n_sim':range(len(sims))})
                for vds in ds.variables:
                    if vds == 'time': continue
                    out[vds] = (('time', 'n_sim',), np.nan*np.zeros((len(out.time), len(out.n_sim))))
                    out[vds].loc[dict(n_sim=c)] = ds[vds]

            else:
                ds = s.Load(vns=vns, decode_times=False)

                # fill output xarray.Dataset
                for vds in ds.variables:
                    if vds == 'time': continue
                    out[vds].loc[dict(n_sim=c)] = ds[vds]

        # optional decode times to cftime 
        if decode_times:
            out = xr.decode_cf(out, use_cftime=use_cftime)

        return out


    # SPECIAL PLOTS
    # TODO: DELETE ?

    def Load_AWTs_DWTs_Plots_sim(self, n_sim=0):
        'Load data needed for WT-WT Probs plot'

        # simulated
        xds_AWT = self.Load_SST_AWT_sim()
        xds_DWT = self.Load_ESTELA_DWT_sim()

        # AWT simulated - evbmus_sims -1 
        xds_AWT = xr.Dataset(
            {'bmus': (('time',), xds_AWT.evbmus_sims.isel(n_sim=n_sim)-1)},
            coords = {'time': xds_AWT.time.values[:]}
        )

        # DWT simulated - evbmus_sims -1
        xds_DWT = xr.Dataset(
            {'bmus': (('time',), xds_DWT.evbmus_sims.isel(n_sim=n_sim)-1)},
            coords = {'time': xds_DWT.time.values[:]}
        )

        # reindex AWT to daily dates (year pad to days)
        xds_AWT = xds_reindex_daily(xds_AWT)

        # get common dates
        dc = xds_common_dates_daily([xds_AWT, xds_DWT])
        xds_DWT = xds_DWT.sel(time=slice(dc[0], dc[-1]))
        xds_AWT = xds_AWT.sel(time=slice(dc[0], dc[-1]))

        return xds_AWT, xds_DWT

    def Load_AWTs_DWTs_Plots_hist(self):
        'Load data needed for WT-WT Probs plot'

        # historical
        xds_AWT = self.Load_SST_KMA()
        xds_DWT = self.Load_ESTELA_KMA()

        # AWT historical - bmus
        xds_AWT = xr.Dataset(
            {'bmus': (('time',), xds_AWT.bmus.values[:])},
            coords = {'time': xds_AWT.time.values[:]}
        )

        # DWT historical - sorted_bmus_storms
        xds_DWT = xr.Dataset(
            {'bmus': (('time',), xds_DWT.sorted_bmus_storms.values[:])},
            coords = {'time': xds_DWT.time.values[:]}
        )

        # reindex AWT to daily dates (year pad to days)
        xds_AWT = xds_reindex_daily(xds_AWT)

        # get common dates
        dc = xds_common_dates_daily([xds_AWT, xds_DWT])
        xds_DWT = xds_DWT.sel(time=slice(dc[0], dc[-1]))
        xds_AWT = xds_AWT.sel(time=slice(dc[0], dc[-1]))

        return xds_AWT, xds_DWT

    def Load_MJO_DWTs_Plots_hist(self):
        'Load data needed for WT-WT Probs plot'

        # historical
        xds_MJO = self.Load_MJO_hist()
        xds_DWT = self.Load_ESTELA_KMA()

        # MJO historical - phase -1
        xds_MJO['phase'] = xds_MJO.phase - 1

        # DWT historical - sorted_bmus
        xds_DWT = xr.Dataset(
            {'bmus': (('time',), xds_DWT.sorted_bmus.values[:])},
            coords = {'time': xds_DWT.time.values[:]}
        )

        # get common dates
        dc = xds_common_dates_daily([xds_DWT, xds_MJO])
        xds_DWT = xds_DWT.sel(time=slice(dc[0], dc[-1]))
        xds_MJO = xds_MJO.sel(time=slice(dc[0], dc[-1]))

        return xds_MJO, xds_DWT

    def Load_MJO_DWTs_Plots_sim(self, n_sim=0):
        'Load data needed for WT-WT Probs plot'

        # simulated
        xds_MJO = self.Load_MJO_sim()
        xds_DWT = self.Load_ESTELA_DWT_sim()

        # DWT simulated - evbmus_sims
        xds_DWT = xr.Dataset(
            {'bmus': (('time',), xds_DWT.evbmus_sims.isel(n_sim=n_sim))},
            coords = {'time': xds_DWT.time.values[:]}
        )

        # get common dates
        dc = xds_common_dates_daily([xds_DWT, xds_MJO])
        xds_DWT = xds_DWT.sel(time=slice(dc[0], dc[-1]))
        xds_MJO = xds_MJO.sel(time=slice(dc[0], dc[-1]))

        return xds_MJO, xds_DWT


    # NEARSHORE

    def Save_HIST_NEARSHORE(self, xds):
        ps = self.paths.site.HISTORICAL.nearshore

        s =  SplitStorage(ps)
        s.Save(xds)

    def Load_HIST_NEARSHORE(self, vns=[], decode_times=False, use_cftime=False):
        ps = self.paths.site.HISTORICAL.nearshore

        s =  SplitStorage(ps)
        return s.Load(vns=vns, decode_times=decode_times, use_cftime=use_cftime)

    def Save_SIM_NEARSHORE(self, xds, n_sim):
        ps = self.paths.site.SIMULATION.nearshore

        nm = '{0:08d}'.format(n_sim)  # sim code
        ps_sim = op.join(ps, nm)

        s =  SplitStorage(ps_sim)
        s.Save(xds, safe_time=True)

    def Load_SIM_NEARSHORE(self, n_sim, vns=[], decode_times=False, use_cftime=False):
        ps = self.paths.site.SIMULATION.nearshore

        nm = '{0:08d}'.format(n_sim)  # sim code
        ps_sim = op.join(ps, nm)

        s =  SplitStorage(ps_sim)
        return s.Load(vns=vns, decode_times=decode_times, use_cftime=use_cftime)

    def Load_SIM_NEARSHORE_all(self, vns=[], decode_times=False, use_cftime=False):
        ps = self.paths.site.SIMULATION.nearshore

        # locate simulations
        sims = sorted([x for x in os.listdir(ps) if x.isdigit() and op.isdir(op.join(ps, x))])

        # read all simulations
        for c, nm in enumerate(sims):
            ps_sim = op.join(ps, nm)
            s = SplitStorage(ps_sim)

            # read time and variables from first file
            if c==0:
                ds = s.Load(vns=vns, decode_times=False)

                # generate output xarray.Dataset, dims and vars
                out = xr.Dataset({}, coords={'time': ds.time, 'n_sim':range(len(sims))})
                for vds in ds.variables:
                    if vds == 'time': continue
                    out[vds] = (('time', 'n_sim',), np.nan*np.zeros((len(out.time), len(out.n_sim))))
                    out[vds].loc[dict(n_sim=c)] = ds[vds]

            else:
                ds = s.Load(vns=vns, decode_times=False)

                # fill output xarray.Dataset
                for vds in ds.variables:
                    if vds == 'time': continue
                    out[vds].loc[dict(n_sim=c)] = ds[vds]

        # optional decode times to cftime 
        if decode_times:
            out = xr.decode_cf(out, use_cftime=use_cftime)

        return out

    def Load_AWAC_buoy(self):

        # buoy path and files
        p_buoy = self.paths.site.HYSWAN.buoy
        fs = ['Roi_AWAC_deploy1.mat', 'Roi_AWAC_deploy2.mat', 'Roi_AWAC_deploy3.mat']

        # aux.
        def Read_AWAC_mat(p_mat):
            'Read a Roi_AWAC_deploy.mat file'

            # matlab file
            mf = ReadMatfile(p_mat)['AWAC']

            # date string format
            str_format = '%d-%b-%Y %H:%M:%S GMT'

            # read times and generate array
            t0 = np.datetime64(datetime.strptime(mf['start_time'], str_format))
            t1 = np.datetime64(datetime.strptime(mf['end_time'], str_format))
            dt = np.timedelta64(mf['profile_interval_s'], 's')

            times = np.arange(t0, t1+dt, dt)

            # variables
            return xr.Dataset(
                {
                    'Hs': (('time',), mf['hs'][0][:len(times)]),
                    'Tp': (('time',), mf['tp'][0][:len(times)]),
                    'Dir': (('time',), mf['dp'][0][:len(times)]),
                },
                coords = {'time':times}
            )

        # read buoy files 
        lx = []
        for f in fs:
            xds = Read_AWAC_mat(op.join(p_buoy, f))
            lx.append(xds)
        xout = xr.combine_by_coords(lx)

        return xout


    # HYCREWW

    def Load_HYCREWW(self):
        'Load RBF coefficients and hycreww sims. max and min values'

        p_h = self.paths.site.HYCREWW.rbf_coef  # RBF_coefficients folder

        # load max and min limits
        p_max = op.join(p_h, 'Max_from_simulations.mat')
        p_min = op.join(p_h, 'Min_from_simulations.mat')
        vs = ['level', 'hs', 'tp', 'rslope', 'bslope', 'rwidth', 'cf']

        smax = ReadMatfile(p_max)['maximum']
        smin = ReadMatfile(p_min)['minimum']
        dl = np.column_stack([smin, smax])

        var_lims = {}
        for c, vn in enumerate(vs):
            var_lims[vn] = dl[c]
        var_lims['hs_lo2'] = [0.005, 0.05]

        # RBF coefficients (for 15 cases)
        code = 'Coeffs_Runup_Xbeach_test'
        n_cases = 15

        RBF_coeffs = []
        for i in range(n_cases):
            cf = ReadMatfile(op.join(p_h, '{0}{1}.mat'.format(code,i+1)))

            rbf = {
                'constant': cf['coeffsave']['RBFConstant'],
                'coeff': cf['coeffsave']['rbfcoeff'],
                'nodes': cf['coeffsave']['x'],
            }

            RBF_coeffs.append(rbf)

        return var_lims, RBF_coeffs


    # CLIMATE CHANGE

    def Load_RCP85(self):
        '''
        Load RCP85 longitude, latitude and ocurrence

        # lon -> (0.5 , 359.5) dlon = 1º
        # lat -> (-89.5, 89.5) dlat = 1º
        '''

        # matfile
        p_mat = self.paths.site.CLIMATE_CHANGE.s4_rcp85
        mf = ReadMatfile(p_mat)

        # data
        lon = mf['lon']
        lat = mf['lat']
        RCP85ratioHIST_occurrence = mf['RCP85ratioHIST_occurrence']

        return lon, lat, RCP85ratioHIST_occurrence

    def Load_SeaLevelRise(self):
        return xr.open_dataset(self.paths.site.CLIMATE_CHANGE.slr_nc, decode_times=True)



# teslakit database aux. objects

class SplitStorage(object):
    'Handles teslakit hourly data split in multiple netCDF4 files'

    def __init__(self, p_storage):

        self.p_storage = p_storage  # main storage folder

    def Save(self, xds, safe_time=False):

        ps = self.p_storage
        if not op.isdir(ps): os.makedirs(ps)

        for vn in xds.variables:
            if vn != 'time':
                save_nc(
                    xds[[vn]], op.join(ps, '{0}.nc'.format(vn)),
                    safe_time=safe_time
                )

    def Load(self, vns=[], decode_times=False, use_cftime=False):

        # aux netCDF4
        def read_var_nc(p, vn):

            # file io
            nc_fid = nc.Dataset(p, 'r')

            # get variable attributes and values
            v_id = nc_fid.variables[vn]
            v_a = dict([(k, v_id.getncattr(k)) for k in v_id.ncattrs()])
            v_v = v_id[:]

            # close file
            nc_fid.close()

            return v_v, v_a

        # storage
        ps = self.p_storage

        # find files to load
        if vns == []:
            fs = glob.glob(op.join(ps,'*.nc'))
        else:
            fs = [op.join(ps, '{0}.nc'.format(v)) for v in vns]
            fs = [f for f in fs if op.isfile(f)]

        # get time (raw) from first file
        v_v_t, v_a_t = read_var_nc(fs[0], 'time')
        v_v_t = v_v_t.astype(int)

        # set calendar 
        #if not 'calendar' in v_a_t.keys():
        #    v_a_t['calendar']='standard'
        v_a_t['calendar']='standard'

        # generate output xarray.Dataset
        out = xr.Dataset({}, coords={'time': v_v_t})
        out['time'].attrs = v_a_t

        # add each file variable
        for f in fs:

            # variable name
            vn = op.basename(f).replace('.nc','')

            # read values and attributes
            v_v, v_a = read_var_nc(f, vn)

            # append to xarray.Dataset
            out[vn] = (('time',), v_v)
            out[vn].attrs = v_a

        # optional decode times to np.datetime64 or DatetimeGregorian
        if decode_times:
            out = xr.decode_cf(out, use_cftime=use_cftime)

        return out


class PathControl(object):
    'auxiliar object for handling teslakit files paths'

    def __init__(self, p_data):

        # ini templates
        p_resources = op.join(op.dirname(op.realpath(__file__)), 'resources')

        # teslakit sites
        p_sites = op.join(p_data, 'sites')

        # string paths
        self.p_data = p_data
        self.p_resources = p_resources
        self.p_sites = p_sites

        self.p_site = None  # current site

        # atdicts
        self.site = None

    def __str__(self):
        'Print paths'
        # TODO: update

        txt = ''
        if isinstance(self.site, dict):
            txt+= '\n\nSite Files:'
            for k1 in sorted(self.site.keys()):
                for k2 in sorted(self.site[k1].keys()):
                    aux1 ='\n.site.{0}.{1}'.format(k1, k2)
                    aux2 = self.site[k1][k2]
                    if 'teslakit' in aux2:
                        aux2 = aux2.split('teslakit')[1]
                    txt+='{0:.<45} {1}'.format(aux1, aux2)
        return txt

    def FilesSummary(self):
        'return detailed database and site files summary'
        # TODO: update

        txt = ''
        if isinstance(self.site,dict):
            txt+= '\n\nSite Files:'
            for k1 in sorted(self.site.keys()):
                for k2 in sorted(self.site[k1].keys()):
                    txt += description(self.site[k1][k2])

        return txt

    def SetSite(self, site_name):
        'Sets dictionary with site files'

        # site folder 
        p_site = op.join(self.p_sites, site_name)
        p_ini = op.join(p_site, 'site.ini')

        # use configparser lib
        cfg = configparser.ConfigParser()
        cfg.read(p_ini)

        dd = {}
        for k in cfg.sections():
            ss = atdict(cfg[k])
            for kk in ss.keys():
                if not ss[kk].startswith(os.sep) and not ':\\' in ss[kk]:
                    ss[kk] = op.join(p_site, k, ss[kk])
            dd[k] = atdict(ss)

        self.site = atdict(dd)
        self.p_site = p_site


class hyswan_db(object):
    '''
    database for MDA-SWAN-RBF methodology
    '''

    def __init__(self, p_base):

        self.p_base = p_base
        self.paths = {
            'sea_dataset': op.join(p_base, 'sea_dataset.pickle'),
            'sea_subset': op.join(p_base, 'sea_subset.pickle'),
            'sea_target': op.join(p_base, 'sea_target.pickle'),

            'swl_dataset': op.join(p_base, 'swl_dataset.pickle'),
            'swl_subset': op.join(p_base, 'swl_subset.pickle'),
            'swl_target': op.join(p_base, 'swl_target.pickle'),

            'swan_projects': op.join(p_base, 'swan_projects'),

            'reconstruction': op.join(p_base, 'recon.pickle'),

        }

    def __str__(self):
        t=''
        for k in sorted(self.paths.keys()):
            t+='{0} - {1}\n'.format(k, self.paths[k])
        return t

    def Save(self, key, pd_data):
        if not op.isdir(self.p_base): os.makedirs(self.p_base)
        pd_data.to_pickle(self.paths[key])

    def Load(self, key):
        return pd.read_pickle(self.paths[key])




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import os.path as op
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
#from mpl_toolkits.basemap import Basemap
from datetime import datetime, timedelta

def WorldMap(
    bk='simple',
    lon_1=-180, lat_1=-70,
    lon_2=180, lat_2=70,
    d_parallel=10., d_meridian=10.):
    'Returns customized world map (mercator)'

    # basemap: mercator
    m = Basemap(
        projection='merc',
        llcrnrlat=lat_1, urcrnrlat=lat_2,
        llcrnrlon=lon_1, urcrnrlon=lon_2,
        lat_ts=20,
        resolution='c')

    # draw parallels.
    parallels = np.arange(-90.,90.,d_parallel)
    m.drawparallels(parallels, labels=[1,0,0,0],fontsize=6)
    # draw meridians
    meridians = np.arange(0.,360.,d_meridian)
    m.drawmeridians(meridians, labels=[0,0,0,1],fontsize=6)

    # draw background
    if bk == 'simple':
        m.drawcoastlines()
        m.fillcontinents(color='coral',lake_color='aqua')
        m.drawmapboundary(fill_color='aqua')
    elif bk == 'shaderelief':
        m.shadedrelief()
    elif bk == 'etopo':
        m.etopo()
    elif bk == 'bluemarble':
        m.bluemarble()

    # return basemap
    return m


def WorldMap_Stations(
    xds_stations, bk='simple',
    lon_1=-180, lat_1=-70,
    lon_2=180, lat_2=70,
    d_parallel=10., d_meridian=10.,
    p_export=None):
    '''
    Plot mercator world map with CSIRO spec stations
    bk (background)= 'simple', 'shaderelief', 'etopo', 'bluemarble'
    lonlat1, lonlat2: basemap limits
    d_parallel, d_meridian: parallels and meridians interval
    '''

    # xds_station lon lat
    lon = xds_stations.longitude.values[:]
    lat = xds_stations.latitude.values[:]
    nms = xds_stations.station_name[:]

    # plot figure
    fig = plt.figure(figsize=(16,9))

    # Get customized basemap
    m = WorldMap(
        bk,
        lon_1, lat_1,
        lon_2, lat_2,
        d_parallel, d_meridian)

    # add stations
    m.scatter(lon, lat, s=24, c='r', latlon=True)

    # more info
    plt.title('CSIRO spec stations')

    # show / export
    if not p_export:
        plt.show()

    else:
        fig.savefig(p_export, dpi=300)
        plt.close()

def WorldGlobe_Stations(
    xds_stations, bk='simple',
    lon_0=0, lat_0=0,
    lon_1=None, lat_1=None,
    lon_2=None, lat_2=None,
    p_export=None):
    '''
    Plot orthogonal world globe with CSIRO spec stations
    bk (background): 'simple', 'shaderelief', 'etopo', 'bluemarble'
    lonlat0: center focus
    lonlat1, lonlat2: basemap limits
    '''
    # TODO: area zoom not working

    # xds_station lon lat
    lon = xds_stations.longitude.values[:]
    lat = xds_stations.latitude.values[:]
    nms = xds_stations.station_name[:]

    # plot figure
    fig = plt.figure(figsize=(12,12))

    # basemap: ortho
    m = Basemap(
        projection='ortho',
        lon_0=lon_0,
        lat_0=lat_0,
        lon_1=lon_1,
        lat_1=lat_1,
        lon_2=lon_2,
        lat_2=lat_2,
        resolution='l',
    )

    # draw parallels.
    parallels = np.arange(-90.,90.,10.)
    m.drawparallels(parallels, labels=[1,0,0,0],fontsize=6)
    # draw meridians
    meridians = np.arange(0.,360.,10.)
    m.drawmeridians(meridians, labels=[0,0,0,1],fontsize=6)

    # draw background
    if bk == 'simple':
        m.drawcoastlines()
        m.fillcontinents(color='coral',lake_color='aqua')
        m.drawmapboundary(fill_color='aqua')
    elif bk == 'shaderelief':
        m.shadedrelief()
    elif bk == 'etopo':
        m.etopo()
    elif bk == 'bluemarble':
        m.bluemarble()

    # add stations
    m.scatter(lon, lat, s=6, c='r', latlon=True)

    # more info
    plt.title('CSIRO spec stations')

    # show / export
    if not p_export:
        plt.show()

    else:
        fig.savefig(p_export, dpi=96)
        plt.close()

def WorldGlobeZoom_Stations(xds_stations, bk='simple', lon0=0, lat0=0, p_export=None):
    '''
    Plot orthogonal world globe zoom with CSIRO spec stations
    bk (background): 'simple', 'shaderelief', 'etopo', 'bluemarble'
    lonlat0: center focus, tuple (lon0, lat0)
    '''
    # TODO: no sale bien

    delta_x = 5000000
    delta_y = 5000000

    # xds_station lon lat
    lon = xds_stations.longitude.values[:]
    lat = xds_stations.latitude.values[:]
    nms = xds_stations.station_name[:]

    # plot figure
    fig = plt.figure(figsize=(12,12))

    # basemap: ortho
    m1 = Basemap(
            projection='ortho',
            lon_0=lon0,
            lat_0=lat0,
            resolution=None
        )
    # zoom
    #ax = fig.add_axes([0.1,0.1,0.8,0.8],facecolor='k')
    xc,yc = m1(lon0, lat0)
    m = Basemap(
        projection='ortho',
        lon_0=lon0,
        lat_0=lat0,
        resolution='l',
        llcrnrx=xc-delta_x, llcrnry=yc-delta_y,
        urcrnrx=yc+delta_x, urcrnry=yc+delta_y,
    )

    # draw parallels.
    parallels = np.arange(-90.,90.,10.)
    # m1.drawparallels(parallels, labels=[1,0,0,0],fontsize=6)
    # draw meridians
    meridians = np.arange(0.,360.,10.)
    #m1.drawmeridians(meridians, labels=[0,0,0,1],fontsize=6)

    # draw background
    if bk == 'simple':
        m.drawcoastlines()
        m.fillcontinents(color='coral',lake_color='aqua')
        m.drawmapboundary(fill_color='aqua')
    elif bk == 'shaderelief':
        m.shadedrelief()
    elif bk == 'etopo':
        m.etopo()
    elif bk == 'bluemarble':
        m.bluemarble()

    # add stations
    m.scatter(lon, lat, s=6, c='r', latlon=True)

    # more info
    plt.title('CSIRO spec stations')

    # show / export
    if not p_export:
        plt.show()

    else:
        fig.savefig(p_export, dpi=96)
        plt.close()


def WorldMap_GriddedCoords(
    xds_gridded, bk='simple',
    lon_1=-180, lat_1=-70,
    lon_2=180, lat_2=70,
    d_parallel=10.,d_meridian=10.,
    p_export=None):
    '''
    Plot mercator world map with CSIRO gridded coordinates
    bk (background)= 'simple', 'shaderelief', 'etopo', 'bluemarble'
    '''

    # xds_station lon lat
    lon = xds_gridded.longitude.values[:]
    lat = xds_gridded.latitude.values[:]

    # land mask 
    use_mask = 'mask' in xds_gridded.variables

    # fix lon for plotting
    lon[np.where(lon>180.0)] = lon[np.where(lon>180.0)]-360

    # plot figure
    fig = plt.figure(figsize=(16,9))

    # Get customized basemap
    m = WorldMap(
        bk,
        lon_1, lat_1,
        lon_2, lat_2,
        d_parallel, d_meridian)

    # add stations
    xx,yy = np.meshgrid(lon, lat)
    if use_mask:
        mask = xds_gridded.mask.values[:]
        xx_plot = xx[mask]
        yy_plot = yy[mask]
    else:
        xx_plot = xx
        yy_plot = yy

    m.scatter(xx_plot, yy_plot, s=6, c='r', latlon=True)

    # more info
    plt.title('CSIRO gridded coords')

    # show / export
    if not p_export:
        plt.show()

    else:
        fig.savefig(p_export, dpi=96)
        plt.close()





from datetime import datetime

import numpy as np
import pandas as pd
import xarray as xr

from ..core.geo import geo_distance_cartesian, geodesic_distance

"""
Dynamic Holland Model for Wind Vortex Fields
This module implements the Dynamic Holland Model to generate wind vortex fields
from storm track parameters. It computes wind speed and direction based on
the storm's position, pressure, and wind parameters, using either spherical or
cartesian coordinates.
The model is optimized for vectorized operations to enhance performance.
It supports both spherical coordinates (latitude, longitude) and cartesian
coordinates (x, y) for storm track data.
The output is an xarray Dataset containing wind speed and direction.
The model is based on the Dynamic Holland Model, which uses storm parameters
to compute wind fields, considering factors like the Coriolis effect,
central pressure deficit, and the radius of maximum winds.
This implementation is designed to be efficient and scalable, suitable for
large datasets and real-time applications.
"""


def vortex_model_grid(
    storm_track: pd.DataFrame,
    cg_lon: np.ndarray,
    cg_lat: np.ndarray,
    coords_mode: str = "SPHERICAL",
) -> xr.Dataset:
    """
    Generate wind vortex fields from storm track parameters using the Dynamic Holland Model.

    Parameters
    ----------
    storm_track : pd.DataFrame
        DataFrame containing storm track parameters.
        - obligatory fields: vfx, vfy, p0, pn, vmax, rmw
        - for SPHERICAL coordinates: lon, lat
        - for CARTESIAN coordinates: x, y, latitude
    cg_lon : np.ndarray
        Computational grid longitudes.
    cg_lat : np.ndarray
        Computational grid latitudes.
    cg_lon, cg_lat : np.ndarray
        Computational grid in longitudes and latitudes.
    coords_mode : str
        'SPHERICAL' for spherical coordinates (latitude, longitude),
        'CARTESIAN' for Cartesian coordinates (x, y).

    Returns
    -------
    xarray.Dataset
        Dataset containing wind speed W, direction Dir (º from north),
        and pressure p at each grid point.

    Examples
    --------
    >>> storm_track = pd.DataFrame({
    ...     'vfx': [10, 12], 'vfy': [5, 6],
    ...     'p0': [1000, 990], 'pn': [980, 970],
    ...     'vmax': [50, 55], 'rmw': [30, 35],
    ...     'lon': [10, 12], 'lat': [20, 22]
    ... })
    >>> cg_lon = np.array([10, 11, 12])
    >>> cg_lat = np.array([20, 21, 22])
    >>> coords_mode = 'SPHERICAL'
    >>> result = vortex_model_grid(storm_track, cg_lon, cg_lat, coords_mode)
    >>> print(result)
    <xarray.Dataset>
    Dimensions:  (lat: 3, lon: 3, time: 2)
    Coordinates:
    * lat      (lat) float64 20.0 21.0 22.0
    * lon      (lon) float64 10.0 11.0 12.0
    * time     (time) datetime64[ns] 2023-10-01 2023-10-02
    Data variables:
    W        (lat, lon, time) float64 0.0 0.0 0.0 ... 0.0 0.0 0.0
    Dir      (lat, lon, time) float64 0.0 0.0 0.0 ... 0.0 0.0 0.0
    p        (lat, lon, time) float64 0.0 0.0 0.0 ... 0.0 0.0 0.0
    """

    # Convert negative longitudes to 0-360 range
    converted_coords = False
    if coords_mode == "SPHERICAL":
        cg_lon = np.where(cg_lon < 0, cg_lon + 360, cg_lon)
        converted_coords = True

    # Define model constants
    beta = 0.9  # Conversion factor of wind speed
    rho_air = 1.15  # Air density [kg/m³]
    omega = 2 * np.pi / 86184.2  # Earth's rotation rate [rad/s]
    deg2rad = np.pi / 180  # Degrees to radians conversion
    conv_1min_to_10min = 0.93  # Convert 1-min avg winds to 10-min avg

    # Extract storm parameters from the DataFrame
    vfx, vfy = storm_track.vfx.values, storm_track.vfy.values  # [kt] translation
    p0, pn = storm_track.p0.values, storm_track.pn.values  # [mbar] pressure
    vmax, rmw = storm_track.vmax.values, storm_track.rmw.values  # [kt], [nmile]
    times = storm_track.index  # time values

    # Select coordinates depending on mode
    if coords_mode == "SPHERICAL":
        x, y, lat = (
            storm_track.lon.values,
            storm_track.lat.values,
            storm_track.lat.values,
        )
    else:
        x, y, lat = (
            storm_track.x.values,
            storm_track.y.values,
            storm_track.latitude.values,
        )

    # Check if the storm is in the southern hemisphere
    is_southern = np.any(lat < 0)

    # Create 2D meshgrid of computational grid
    lon2d, lat2d = np.meshgrid(cg_lon, cg_lat)
    shape = (len(cg_lat), len(cg_lon), len(p0))

    # Initialize output arrays for wind magnitude and direction
    W = np.zeros(shape)
    D = np.zeros(shape)
    p = np.zeros(shape)

    # Loop over time steps to compute vortex fields
    for i in range(len(p0)):
        lo, la, la0 = x[i], y[i], lat[i]

        # Skip time steps with NaN values
        if np.isnan([lo, la, la0, p0[i], pn[i], vfx[i], vfy[i], vmax[i], rmw[i]]).any():
            continue

        # Compute distance between grid points and storm center
        if coords_mode == "SPHERICAL":
            geo_dis = geodesic_distance(lat2d, lon2d, la, lo)
            RE = 6378.135 * 1000  # Earth radius [m]
            r = geo_dis * np.pi / 180.0 * RE
        elif coords_mode == "CARTESIAN":
            r = geo_distance_cartesian(lat2d, lon2d, la, lo)

        # Compute direction from storm center to each grid point
        dlat = (lat2d - la) * deg2rad
        dlon = (lon2d - lo) * deg2rad
        theta = np.arctan2(dlat, -dlon if is_southern else dlon)

        # Compute central pressure deficit [Pa]
        CPD = max((pn[i] - p0[i]) * 100, 100)

        # Compute Coriolis parameter
        coriolis = 2 * omega * np.sin(abs(la0) * deg2rad)

        # Compute adjusted gradient wind speed
        v_trans = np.hypot(vfx[i], vfy[i])  # [kt] translation magnitude
        vkt = vmax[i] - v_trans  # corrected max wind [kt]
        vgrad = vkt / beta  # gradient wind [kt]
        vm = vgrad * 0.52  # convert to [m/s]

        # Compute radius and nondimensional radius
        rm = rmw[i] * 1.852 * 1000  # [m]
        rn = rm / r  # nondimensional

        # Compute Holland B parameter, with bounds
        B = np.clip(rho_air * np.exp(1) * vm**2 / CPD, 1, 2.5)

        # Gradient wind velocity at each grid point
        vg = (
            np.sqrt((rn**B) * np.exp(1 - rn**B) * vm**2 + (r**2 * coriolis**2) / 4)
            - r * coriolis / 2
        )

        # Convert to wind components at 10m height
        sign = 1 if is_southern else -1
        ve = sign * vg * beta * np.sin(theta) * conv_1min_to_10min
        vn = vg * beta * np.cos(theta) * conv_1min_to_10min

        # Add translation velocity components
        vtae = (np.abs(vg) / vgrad) * vfx[i]
        vtan = (np.abs(vg) / vgrad) * vfy[i]
        vfe = ve + vtae
        vfn = vn + vtan

        # Total wind magnitude [m/s]
        W[:, :, i] = np.hypot(vfe, vfn)

        # Pressure gradient to estimate wind direction
        pr = p0[i] + (pn[i] - p0[i]) * np.exp(-(rn**B))  # surface pressure
        p[:, :, i] = pr
        py, px = np.gradient(pr)
        angle = np.arctan2(py, px) + np.sign(la0) * np.pi / 2

        # Wind direction in degrees from north (clockwise)
        D[:, :, i] = (270 - np.rad2deg(angle)) % 360

    # Define coordinate labels based on coordinate mode
    ylab, xlab = ("lat", "lon") if coords_mode == "SPHERICAL" else ("y", "x")

    if converted_coords:
        # Convert longitudes back to -180 to 180 range if they were converted
        cg_lon = np.where(cg_lon > 180, cg_lon - 360, cg_lon)

    # Return results as xarray Dataset
    return xr.Dataset(
        {
            "W": ((ylab, xlab, "time"), W, {"units": "m/s"}),
            "Dir": ((ylab, xlab, "time"), D, {"units": "º"}),
            "p": ((ylab, xlab, "time"), p * 100, {"units": "Pa"}),
        },
        coords={ylab: cg_lat, xlab: cg_lon, "time": times},
    )


def vortex2delft_3D_FM_nc(
    mesh: xr.Dataset,
    ds_vortex: xr.Dataset,
) -> xr.Dataset:
    """
    Convert the vortex dataset to a Delft3D FM compatible netCDF forcing file.

    Parameters
    ----------
    mesh : xarray.Dataset
        The mesh dataset containing the node coordinates.
    ds_vortex : xarray.Dataset
        The vortex dataset containing wind speed and pressure data.
    path_output : str
        The output path where the netCDF file will be saved.
    ds_name : str
        The name of the output netCDF file, default is "forcing_Tonga_vortex.nc".
    forcing_ext : str
        The extension for the forcing file, default is "GreenSurge_GFDcase_wind.ext".
    Returns
    -------
    xarray.Dataset
        A dataset containing the interpolated wind speed and pressure data,
        ready for use in Delft3D FM.
    """
    longitude = mesh.mesh2d_node_x.values
    latitude = mesh.mesh2d_node_y.values
    n_time = ds_vortex.time.size

    lat_interp = xr.DataArray(latitude, dims="node")
    lon_interp = xr.DataArray(longitude, dims="node")

    angle = np.deg2rad((270 - ds_vortex.Dir.values) % 360)
    W = ds_vortex.W.values

    ds_vortex_interp = xr.Dataset(
        {
            "windx": (
                ("latitude", "longitude", "time"),
                (W * np.cos(angle)).astype(np.float32),
            ),
            "windy": (
                ("latitude", "longitude", "time"),
                (W * np.sin(angle)).astype(np.float32),
            ),
            "airpressure": (
                ("latitude", "longitude", "time"),
                ds_vortex.p.values.astype(np.float32),
            ),
        },
        coords={
            "latitude": ds_vortex.lat.values,
            "longitude": ds_vortex.lon.values,
            "time": np.arange(n_time),
        },
    )

    forcing_dataset = ds_vortex_interp.interp(latitude=lat_interp, longitude=lon_interp)

    reference_date_str = (
        ds_vortex.time.values[0]
        .astype("M8[ms]")
        .astype(datetime)
        .strftime("%Y-%m-%d %H:%M:%S")
    )

    forcing_dataset["windx"].attrs = {
        "coordinates": "time node",
        "long_name": "Wind speed in x direction",
        "standard_name": "windx",
        "units": "m s-1",
    }
    forcing_dataset["windy"].attrs = {
        "coordinates": "time node",
        "long_name": "Wind speed in y direction",
        "standard_name": "windy",
        "units": "m s-1",
    }

    forcing_dataset["airpressure"].attrs = {
        "coordinates": "time node",
        "long_name": "Atmospheric Pressure",
        "standard_name": "air_pressure",
        "units": "Pa",
    }

    forcing_dataset["time"].attrs = {
        "standard_name": "time",
        "long_name": f"Time - hours since {reference_date_str} +00:00",
        "time_origin": f"{reference_date_str}",
        "units": f"hours since {reference_date_str} +00:00",
        "calendar": "gregorian",
        "description": "Time definition for the forcing data",
    }

    forcing_dataset["longitude"].attrs = {
        "description": "Longitude of each mesh node of the computational grid",
        "standard_name": "longitude",
        "long_name": "longitude",
        "units": "degrees_east",
    }
    forcing_dataset["latitude"].attrs = {
        "description": "Latitude of each mesh node of the computational grid",
        "standard_name": "latitude",
        "long_name": "latitude",
        "units": "degrees_north",
    }

    return forcing_dataset



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common
import os
import os.path as op
from datetime import datetime, timedelta

# pip
import numpy as np

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.dates as mdates
import matplotlib.colors as mcolors
from matplotlib.colorbar import ColorbarBase

# teslakit
from ..util.operations import GetBestRowsCols
from .custom_colors import GetClusterColors
from ..kma import ClusterProbabilities, ChangeProbabilities

# import constants
from .config import _faspect, _fsize, _fdpi


def Plot_RCP_ocurrence(lon_grid, lat_grid, ocurrence, site_lon_index,
                       site_lat_index, show=True):
    '''
    Plot global map with RCP TCs ocurrence and study site point

    lon_grid, lat_grid - 2D numpy array. ocurrence coordinates meshgrid
    '''

    # parameters
    ttl = 'Tropical Cyclone tracks: [prob.RCP85/prob.HIST - 1] (%)'
    ttl += '\nProjected change in TCs occurrence rate (%)'

    # site location
    lon_site = lon_grid[site_lon_index, site_lat_index]
    lat_site = lat_grid[site_lon_index, site_lat_index]

    # plot figure
    fig, ax = plt.subplots(figsize=(_faspect*_fsize, _fsize))

    # Plot global map and location of Site
    pc = ax.pcolor(lon_grid, lat_grid, ocurrence, cmap='bwr', vmin=-40, vmax=40, label='')
    ax.plot(lon_site, lat_site, 'ok',label='Site')

    # colorbar
    plt.colorbar(pc, ax=ax)

    # titles and labels
    ax.set_title(ttl)
    ax.set_xlabel('Longitude (º)')
    ax.set_ylabel('Latitude (º)')

    plt.legend()

    # show and return figure
    if show: plt.show()
    return fig


